"""Wrappers over FieldData gRPC service of Fluent."""
from enum import IntEnum
from functools import reduce
from typing import Callable, Dict, List, Optional, Tuple, Union

import grpc
import numpy as np

from ansys.api.fluent.v0 import field_data_pb2 as FieldDataProtoModule
from ansys.api.fluent.v0 import field_data_pb2_grpc as FieldGrpcModule
from ansys.fluent.core.services.error_handler import catch_grpc_error
from ansys.fluent.core.services.interceptors import BatchInterceptor, TracingInterceptor
from ansys.fluent.core.solver.error_message import allowed_name_error_message


def override_help_text(func, func_to_be_wrapped):
    """Override function help text."""
    func.__doc__ = "\n" + func_to_be_wrapped.__doc__
    func.__name__ = func_to_be_wrapped.__qualname__
    return func


# this can be switched to False in scenarios where the field_data request inputs are
# fed by results of field_info queries, which might be true in GUI code.
validate_inputs = True


class FieldDataService:
    """FieldData service of Fluent."""

    def __init__(self, channel: grpc.Channel, metadata):
        """__init__ method of FieldDataService class."""
        intercept_channel = grpc.intercept_channel(
            channel, TracingInterceptor(), BatchInterceptor()
        )
        self.__stub = FieldGrpcModule.FieldDataStub(intercept_channel)
        self.__metadata = metadata

    @catch_grpc_error
    def get_scalar_fields_range(self, request):
        """GetRange rpc of FieldData service."""
        return self.__stub.GetRange(request, metadata=self.__metadata)

    @catch_grpc_error
    def get_scalar_fields_info(self, request):
        """GetFieldsInfo rpc of FieldData service."""
        return self.__stub.GetFieldsInfo(request, metadata=self.__metadata)

    @catch_grpc_error
    def get_vector_fields_info(self, request):
        """GetVectorFieldsInfo rpc of FieldData service."""
        return self.__stub.GetVectorFieldsInfo(request, metadata=self.__metadata)

    @catch_grpc_error
    def get_surfaces_info(self, request):
        """GetSurfacesInfo rpc of FieldData service."""
        return self.__stub.GetSurfacesInfo(request, metadata=self.__metadata)

    @catch_grpc_error
    def get_fields(self, request):
        """GetFields rpc of FieldData service."""
        return self.__stub.GetFields(request, metadata=self.__metadata)


class FieldInfo:
    """Provides access to Fluent field information.

    Methods
    -------
    get_scalar_fields_range(field: str, node_value: bool, surface_ids: List[int])
    -> List[float]
        Get the range (minimum and maximum values) of the field.

    get_scalar_fields_info(self) -> dict
        Get fields information (field name, domain, and section).

    get_vector_fields_info(self) -> dict
        Get vector fields information.

    get_surfaces_info(self) -> dict
        Get surfaces information (surface name, ID, and type).
    """

    def __init__(self, service: FieldDataService):
        """__init__ method of FieldInfo class."""
        self._service = service

    def get_scalar_fields_range(
        self, field: str, node_value: bool = False, surface_ids: List[int] = None
    ) -> List[float]:
        """Get the range (minimum and maximum values) of the field.

        Parameters
        ----------
        field: str
            Name of the field
        node_value: bool
        surface_ids : List[int], optional
            List of surface IDS for the surface data.

        Returns
        -------
        List[float]
        """
        if not surface_ids:
            surface_ids = []
        request = FieldDataProtoModule.GetRangeRequest()
        request.fieldName = field
        request.nodeValue = node_value
        request.surfaceid.extend(
            [FieldDataProtoModule.SurfaceId(id=int(id)) for id in surface_ids]
        )
        response = self._service.get_scalar_fields_range(request)
        return [response.minimum, response.maximum]

    def get_scalar_fields_info(self) -> dict:
        """Get fields information (field name, domain, and section).

        Returns
        -------
        Dict
        """
        request = FieldDataProtoModule.GetFieldsInfoRequest()
        response = self._service.get_scalar_fields_info(request)
        return {
            field_info.solverName: {
                "display_name": field_info.displayName,
                "section": field_info.section,
                "domain": field_info.domain,
            }
            for field_info in response.fieldInfo
        }

    def get_vector_fields_info(self) -> dict:
        """Get vector fields information (vector components).

        Returns
        -------
        Dict
        """
        request = FieldDataProtoModule.GetVectorFieldsInfoRequest()
        response = self._service.get_vector_fields_info(request)
        return {
            vector_field_info.displayName: {
                "x-component": vector_field_info.xComponent,
                "y-component": vector_field_info.yComponent,
                "z-component": vector_field_info.zComponent,
            }
            for vector_field_info in response.vectorFieldInfo
        }

    def get_surfaces_info(self) -> dict:
        """Get surfaces information (surface name, ID, and type).

        Returns
        -------
        Dict
        """
        request = FieldDataProtoModule.GetSurfacesInfoResponse()
        response = self._service.get_surfaces_info(request)
        info = {
            surface_info.surfaceName: {
                "surface_id": [surf.id for surf in surface_info.surfaceId],
                "zone_id": surface_info.zoneId.id,
                "zone_type": surface_info.zoneType,
                "type": surface_info.type,
            }
            for surface_info in response.surfaceInfo
        }
        return info


def unavailable_field_error_message(context: str, field_name: str) -> str:
    """Error message for unavailable fields."""
    return f"{field_name} is not a currently available {context}."


class FieldNameError(ValueError):
    """Exception class for errors in field name."""

    pass


class ScalarFieldNameError(FieldNameError):
    """Exception class for errors in scalar field name."""

    def __init__(self, field_name: str, allowed_values: List[str]):
        """__init__ method of ScalarFieldNameError class."""
        self.field_name = field_name
        super().__init__(
            allowed_name_error_message("scalar field", field_name, allowed_values)
        )


class VectorFieldNameError(FieldNameError):
    """Exception class for errors in vector field name."""

    def __init__(self, field_name: str, allowed_values: List[str]):
        """__init__ method of VectorFieldNameError class."""
        self.field_name = field_name
        super().__init__(
            allowed_name_error_message("vector field", field_name, allowed_values)
        )


class FieldUnavailable(RuntimeError):
    """Exception class for when field is unavailable."""

    pass


class ScalarFieldUnavailable(FieldUnavailable):
    """Exception class for when scalar field is unavailable."""

    def __init__(self, field_name: str):
        """__init__ method of ScalarFieldUnavailable class."""
        self.field_name = field_name
        super().__init__(unavailable_field_error_message("scalar field", field_name))


class VectorFieldUnavailable(FieldUnavailable):
    """Exception class for when vector field is unavailable."""

    def __init__(self, field_name: str):
        """__init__ method of VectorFieldUnavailable class."""
        self.field_name = field_name
        super().__init__(unavailable_field_error_message("vector field", field_name))


class SurfaceNameError(ValueError):
    """Exception class for errors in surface name."""

    def __init__(self, surface_name: str, allowed_values: List[str]):
        """__init__ method of SurfaceNameError class."""
        self.surface_name = surface_name
        super().__init__(
            allowed_name_error_message("surface", surface_name, allowed_values)
        )


class SurfaceDataType(IntEnum):
    """Provides surface data types."""

    Vertices = 1
    FacesConnectivity = 2
    FacesNormal = 3
    FacesCentroid = 4


class _AllowedNames:
    def __init__(self, field_info: FieldInfo):
        self._field_info = field_info

    def is_valid(self, name, respect_data_valid=True):
        return name in self(respect_data_valid)


class _AllowedFieldNames(_AllowedNames):
    def __init__(self, field_info: FieldInfo, is_data_valid: Callable[[], bool]):
        super().__init__(field_info=field_info)
        self._is_data_valid = is_data_valid

    def valid_name(self, field_name):
        if validate_inputs:
            names = self
            if not names.is_valid(field_name, respect_data_valid=False):
                raise self._field_name_error(
                    field_name=field_name,
                    allowed_values=names(respect_data_valid=False),
                )
            if not names.is_valid(field_name, respect_data_valid=True):
                raise self._field_unavailable_error(field_name)
        return field_name


class _AllowedSurfaceNames(_AllowedNames):
    def __call__(self, respect_data_valid: bool = True) -> List[str]:
        return self._field_info.get_surfaces_info()

    def valid_name(self, surface_name: str) -> str:
        if validate_inputs:
            if not self.is_valid(surface_name):
                raise SurfaceNameError(
                    surface_name=surface_name,
                    allowed_values=self(),
                )
        return surface_name


class _AllowedSurfaceIDs(_AllowedNames):
    def __call__(self, respect_data_valid: bool = True) -> List[int]:
        try:
            return [
                info["surface_id"][0]
                for _, info in self._field_info.get_surfaces_info().items()
            ]
        except (KeyError, IndexError):
            pass


class _AllowedScalarFieldNames(_AllowedFieldNames):
    _field_name_error = ScalarFieldNameError
    _field_unavailable_error = ScalarFieldUnavailable

    def __call__(self, respect_data_valid: bool = True) -> List[str]:
        field_dict = self._field_info.get_scalar_fields_info()
        return (
            field_dict
            if (not respect_data_valid or self._is_data_valid())
            else [
                name
                for name, info in field_dict.items()
                if info["section"] in ("Mesh...", "Cell Info...")
            ]
        )


class _AllowedVectorFieldNames(_AllowedFieldNames):
    _field_name_error = VectorFieldNameError
    _field_unavailable_error = VectorFieldUnavailable

    def __call__(self, respect_data_valid: bool = True) -> List[str]:
        return (
            self._field_info.get_vector_fields_info()
            if (not respect_data_valid or self._is_data_valid())
            else []
        )

    def is_valid(self, name, respect_data_valid=True):
        return name in self(respect_data_valid)


class _FieldMethod:
    class _Arg:
        def __init__(self, accessor):
            self._accessor = accessor

        def allowed_values(self):
            return sorted(self._accessor())

    def __init__(self, field_data_accessor, args_allowed_values_accessors):
        self._field_data_accessor = field_data_accessor
        for arg_name, accessor in args_allowed_values_accessors.items():
            setattr(self, arg_name, _FieldMethod._Arg(accessor))

    def __call__(self, *args, **kwargs):
        return self._field_data_accessor(*args, **kwargs)


class FieldTransaction:
    """Populates Fluent field data on surfaces."""

    def __init__(
        self,
        service: FieldDataService,
        field_info: FieldInfo,
        allowed_surface_ids,
        allowed_surface_names,
        allowed_scalar_field_names,
        allowed_vector_field_names,
    ):
        """__init__ method of FieldTransaction class."""
        self._service = service
        self._field_info = field_info
        self._fields_request = get_fields_request()

        self._allowed_surface_names = allowed_surface_names
        self._allowed_scalar_field_names = allowed_scalar_field_names
        self._allowed_vector_field_names = allowed_vector_field_names

        surface_args = dict(
            surface_ids=allowed_surface_ids,
            surface_names=self._allowed_surface_names,
        )
        scalar_field_args = {
            **dict(field_name=self._allowed_scalar_field_names),
            **surface_args,
        }
        self.add_scalar_fields_request = override_help_text(
            _FieldMethod(
                field_data_accessor=self.add_scalar_fields_request,
                args_allowed_values_accessors=scalar_field_args,
            ),
            self.add_scalar_fields_request,
        )
        self.add_vector_fields_request = override_help_text(
            _FieldMethod(
                field_data_accessor=self.add_vector_fields_request,
                args_allowed_values_accessors={
                    **dict(field_name=self._allowed_vector_field_names),
                    **surface_args,
                },
            ),
            self.add_vector_fields_request,
        )
        self.add_surfaces_request = override_help_text(
            _FieldMethod(
                field_data_accessor=self.add_surfaces_request,
                args_allowed_values_accessors=surface_args,
            ),
            self.add_surfaces_request,
        )
        self.add_pathlines_fields_request = override_help_text(
            _FieldMethod(
                field_data_accessor=self.add_pathlines_fields_request,
                args_allowed_values_accessors=scalar_field_args,
            ),
            self.add_pathlines_fields_request,
        )

    def add_surfaces_request(
        self,
        surface_ids: Optional[List[int]] = None,
        surface_names: Optional[List[str]] = None,
        overset_mesh: Optional[bool] = False,
        provide_vertices: Optional[bool] = True,
        provide_faces: Optional[bool] = True,
        provide_faces_centroid: Optional[bool] = False,
        provide_faces_normal: Optional[bool] = False,
    ) -> None:
        """Add request to get surface data (vertices, face connectivity,
        centroids, and normals).

        Parameters
        ----------
        surface_ids : List[int], optional
            List of surface IDS for the surface data.
        surface_names: List[str], optional
            List of surface names for the surface data.
        overset_mesh : bool, optional
            Whether to get the overset met. The default is ``False``.
        provide_vertices : bool, optional
            Whether to get node coordinates. The default is ``True``.
        provide_faces : bool, optional
            Whether to get face connectivity. The default is ``True``.
        provide_faces_centroid : bool, optional
            Whether to get face centroids. The default is ``False``.
        provide_faces_normal : bool, optional
            Whether to get faces normal. The default is ``False``

        Returns
        -------
        None
        """
        surface_ids = _get_surface_ids(
            field_info=self._field_info,
            allowed_surface_names=self._allowed_surface_names,
            surface_ids=surface_ids,
            surface_names=surface_names,
        )
        self._fields_request.surfaceRequest.extend(
            [
                FieldDataProtoModule.SurfaceRequest(
                    surfaceId=surface_id,
                    oversetMesh=overset_mesh,
                    provideFaces=provide_faces,
                    provideVertices=provide_vertices,
                    provideFacesCentroid=provide_faces_centroid,
                    provideFacesNormal=provide_faces_normal,
                )
                for surface_id in surface_ids
            ]
        )

    def add_scalar_fields_request(
        self,
        field_name: str,
        surface_ids: Optional[List[int]] = None,
        surface_names: Optional[List[str]] = None,
        node_value: Optional[bool] = True,
        boundary_value: Optional[bool] = False,
    ) -> None:
        """Add request to get scalar field data on surfaces.

        Parameters
        ----------
        field_name : str
            Name of the scalar field.
        surface_ids : List[int], optional
            List of surface IDs for scalar field data.
        surface_names: List[str], optional
            List of surface names for scalar field data.
        node_value : bool, optional
            Whether to provide the nodal location. The default is ``True``. If
            ``False``, the element location is provided.
        boundary_value : bool, optional
            Whether to provide the slip velocity at the wall boundaries. The default
            is ``False``. When ``True``, no slip velocity is provided.

        Returns
        -------
        None
        """
        surface_ids = _get_surface_ids(
            field_info=self._field_info,
            allowed_surface_names=self._allowed_surface_names,
            surface_ids=surface_ids,
            surface_names=surface_names,
        )
        self._fields_request.scalarFieldRequest.extend(
            [
                FieldDataProtoModule.ScalarFieldRequest(
                    surfaceId=surface_id,
                    scalarFieldName=self._allowed_scalar_field_names.valid_name(
                        field_name
                    ),
                    dataLocation=FieldDataProtoModule.DataLocation.Nodes
                    if node_value
                    else FieldDataProtoModule.DataLocation.Elements,
                    provideBoundaryValues=boundary_value,
                )
                for surface_id in surface_ids
            ]
        )

    def add_vector_fields_request(
        self,
        field_name: str,
        surface_ids: Optional[List[int]] = None,
        surface_names: Optional[List[str]] = None,
    ) -> None:
        """Add request to get vector field data on surfaces.

        Parameters
        ----------
        field_name : str
            Name of the vector field.
        surface_ids : List[int], optional
            List of surface IDs for vector field data.
        surface_names: List[str], optional
            List of surface names for vector field data.

        Returns
        -------
        None
        """
        surface_ids = _get_surface_ids(
            field_info=self._field_info,
            allowed_surface_names=self._allowed_surface_names,
            surface_ids=surface_ids,
            surface_names=surface_names,
        )
        self._fields_request.vectorFieldRequest.extend(
            [
                FieldDataProtoModule.VectorFieldRequest(
                    surfaceId=surface_id,
                    vectorFieldName=self._allowed_vector_field_names.valid_name(
                        field_name,
                    ),
                )
                for surface_id in surface_ids
            ]
        )

    def add_pathlines_fields_request(
        self,
        field_name: str,
        surface_ids: Optional[List[int]] = None,
        surface_names: Optional[List[str]] = None,
        additional_field_name: Optional[str] = "",
        provide_particle_time_field: Optional[bool] = False,
        node_value: Optional[bool] = True,
        steps: Optional[int] = 500,
        step_size: Optional[float] = 0.01,
        skip: Optional[int] = 0,
        reverse: Optional[bool] = False,
        accuracy_control_on: Optional[bool] = False,
        tolerance: Optional[float] = 0.001,
        coarsen: Optional[int] = 1,
        velocity_domain: Optional[str] = "all-phases",
        zones: Optional[list] = [],
    ) -> None:
        """Add request to get pathlines field on surfaces.

        Parameters
        ----------
        field_name : str
            Name of the scalar field to color pathlines.
        surface_ids : List[int], optional
            List of surface IDs for pathlines field data.
        surface_names : List[str], optional
            List of surface names for pathlines field data.
        additional_field_name : str, optional
            Additional field if required.
        provide_particle_time_field: bool, optional
            Whether to provide the particle time. The default is ``False``.
        node_value : bool, optional
                    Whether to provide the nodal values. The default is ``True``. If
                    ``False``, element values are provided.
        steps: int, optional
            Pathlines steps. The default is ``500``
        step_size: float, optional
            Pathlines step size. The default is ``0.01``.
        skip: int, optional
            Pathlines to skip. The default is ``0``.
        reverse: bool, optional
            Whether to draw pathlines in reverse direction. The default is ``False``.
        tolerance: float, optional
            Pathlines tolerance. The default is ``0.001``.
        coarsen: int, optional
            Pathlines coarsen. The default is ``1``.
        velocity_domain: str, optional
            Domain for pathlines. The default is ``"all-phases"``.
        zones: list, optional
            Zones for pathlines. The default is ``[]``.
        Returns
        -------
        None
        """
        surface_ids = _get_surface_ids(
            field_info=self._field_info,
            allowed_surface_names=self._allowed_surface_names,
            surface_ids=surface_ids,
            surface_names=surface_names,
        )
        self._fields_request.pathlinesFieldRequest.extend(
            [
                FieldDataProtoModule.PathlinesFieldRequest(
                    surfaceId=surface_id,
                    field=field_name,
                    additionalField=additional_field_name,
                    provideParticleTimeField=provide_particle_time_field,
                    dataLocation=FieldDataProtoModule.DataLocation.Nodes
                    if node_value
                    else FieldDataProtoModule.DataLocation.Elements,
                    steps=steps,
                    stepSize=step_size,
                    skip=skip,
                    reverse=reverse,
                    accuracyControlOn=accuracy_control_on,
                    tolerance=tolerance,
                    coarsen=coarsen,
                    velocityDomain=velocity_domain,
                    zones=zones,
                )
                for surface_id in surface_ids
            ]
        )

    def get_fields(self) -> Dict[Union[int, Tuple], Dict[int, Dict[str, np.array]]]:
        """Get data for previously added requests and then clear all requests.

        Returns
        -------
        Dict[int, Dict[int, Dict[str, np.array]]]
            Data is returned as dictionary of dictionaries in the following structure:
            tag Union[int, Tuple]-> surface_id [int] -> field_name [str] -> field_data[np.array]

            The tag is a tuple for Fluent 2023 R1 or later.
        """
        return extract_fields(self._service.get_fields(self._fields_request))

    def __call__(self):
        self.get_fields()


class _FieldDataConstants:
    """Defines constants for Fluent field data."""

    # data mapping
    proto_field_type_to_np_data_type = {
        FieldDataProtoModule.FieldType.INT_ARRAY: np.int32,
        FieldDataProtoModule.FieldType.LONG_ARRAY: np.int64,
        FieldDataProtoModule.FieldType.FLOAT_ARRAY: np.float32,
        FieldDataProtoModule.FieldType.DOUBLE_ARRAY: np.float64,
    }
    np_data_type_to_proto_field_type = {
        np.int32: FieldDataProtoModule.FieldType.INT_ARRAY,
        np.int64: FieldDataProtoModule.FieldType.LONG_ARRAY,
        np.float32: FieldDataProtoModule.FieldType.FLOAT_ARRAY,
        np.float64: FieldDataProtoModule.FieldType.DOUBLE_ARRAY,
    }
    chunk_size = 256 * 1024
    bytes_stream = True
    payloadTags = {
        FieldDataProtoModule.PayloadTag.OVERSET_MESH: 1,
        FieldDataProtoModule.PayloadTag.ELEMENT_LOCATION: 2,
        FieldDataProtoModule.PayloadTag.NODE_LOCATION: 4,
        FieldDataProtoModule.PayloadTag.BOUNDARY_VALUES: 8,
    }


def _get_surface_ids(
    field_info: FieldInfo,
    allowed_surface_names,
    surface_ids: Optional[List[int]] = None,
    surface_names: Optional[List[str]] = None,
    surface_name: Optional[str] = None,
) -> List[int]:
    """Get surface ids' based on surface names or ids'.

    Parameters
    ----------
    surface_ids : List[int], optional
        List of surface IDs.
    surface_names: List[str], optional
        List of surface names.
    surface_name: str, optional
        List of surface name.

    Returns
    -------
    List[int]
    """
    if surface_ids and (surface_name or surface_names):
        raise RuntimeError("Please provide either surface names or surface ids.")
    if not surface_ids:
        surface_ids = []
        if surface_names:
            for surface_name in surface_names:
                surface_ids.extend(
                    field_info.get_surfaces_info()[surface_name]["surface_id"]
                )
        elif surface_name:
            surface_ids = field_info.get_surfaces_info()[
                allowed_surface_names.valid_name(surface_name)
            ]["surface_id"]
        else:
            raise RuntimeError("Please provide either surface names or surface ids.")
    return surface_ids


def get_fields_request():
    """Populates a new field request."""
    return FieldDataProtoModule.GetFieldsRequest(
        provideBytesStream=_FieldDataConstants.bytes_stream,
        chunkSize=_FieldDataConstants.chunk_size,
    )


def extract_fields(chunk_iterator):
    """Extracts field data via a server call."""

    def _get_tag_for_surface_request(surface_request):
        return (("type", "surface-data"),)

    def _get_tag_for_vector_field_request(vector_field_request):
        return (("type", "vector-field"),)

    def _get_tag_for_scalar_field_request(scalar_field_request):
        return (
            ("type", "scalar-field"),
            ("dataLocation", scalar_field_request.dataLocation),
            ("boundaryValues", scalar_field_request.provideBoundaryValues),
        )

    def _get_tag_for_pathlines_field_request(pathlines_field_request):
        return (("type", "pathlines-field"), ("field", pathlines_field_request.field))

    def _extract_field(field_datatype, field_size, chunk_iterator):
        if not chunk_iterator.is_active():
            raise RuntimeError(
                "Unexpectedly encountered empty chunk during field extraction."
            )
        field_arr = np.empty(field_size, dtype=field_datatype)
        field_datatype_item_size = np.dtype(field_datatype).itemsize
        index = 0
        for chunk in chunk_iterator:
            if chunk.bytePayload:
                count = min(
                    len(chunk.bytePayload) // field_datatype_item_size,
                    field_size - index,
                )
                field_arr[index : index + count] = np.frombuffer(
                    chunk.bytePayload, field_datatype, count=count
                )
                index += count
                if index == field_size:
                    return field_arr
            else:
                payload = (
                    chunk.floatPayload.payload
                    or chunk.intPayload.payload
                    or chunk.doublePayload.payload
                    or chunk.longPayload.payload
                )
                count = len(payload)
                field_arr[index : index + count] = np.fromiter(
                    payload, dtype=field_datatype
                )
                index += count
                if index == field_size:
                    return field_arr

    fields_data = {}
    for chunk in chunk_iterator:
        payload_info = chunk.payloadInfo
        field = _extract_field(
            _FieldDataConstants.proto_field_type_to_np_data_type[
                payload_info.fieldType
            ],
            payload_info.fieldSize,
            chunk_iterator,
        )

        surface_id = payload_info.surfaceId

        field_request_info = payload_info.fieldRequestInfo
        request_type = field_request_info.WhichOneof("request")
        if request_type is not None:
            payload_tag_id = (
                _get_tag_for_surface_request(field_request_info.surfaceRequest)
                if request_type == "surfaceRequest"
                else _get_tag_for_scalar_field_request(
                    field_request_info.scalarFieldRequest
                )
                if request_type == "scalarFieldRequest"
                else _get_tag_for_vector_field_request(
                    field_request_info.vectorFieldRequest
                )
                if request_type == "vectorFieldRequest"
                else _get_tag_for_pathlines_field_request(
                    field_request_info.pathlinesFieldRequest
                )
                if request_type == "pathlinesFieldRequest"
                else None
            )
        else:
            payload_tag_id = reduce(
                lambda x, y: x | y,
                [
                    _FieldDataConstants.payloadTags[tag]
                    for tag in payload_info.payloadTag
                ]
                or [0],
            )
        payload_data = fields_data.get(payload_tag_id)
        if not payload_data:
            payload_data = fields_data[payload_tag_id] = {}
        surface_data = payload_data.get(surface_id)
        if surface_data:
            if payload_info.fieldName in surface_data:
                surface_data.update(
                    {
                        payload_info.fieldName: np.concatenate(
                            (surface_data[payload_info.fieldName], field)
                        )
                    }
                )
            else:
                surface_data.update({payload_info.fieldName: field})
        else:
            payload_data[surface_id] = {payload_info.fieldName: field}
    return fields_data


class BaseFieldData:
    """Contains common properties required by all field data types."""

    def __init__(self, i_d, data):
        self._data = data
        self._id = i_d

    @property
    def data(self):
        return self._data

    @property
    def surface_id(self):
        return self._id

    @property
    def size(self):
        return len(self._data)

    def __getitem__(self, item):
        return self._data[item]


class ScalarFieldData(BaseFieldData):
    """Contains scalar field data."""

    class ScalarData:
        """Stores and provides the data as a scalar."""

        def __init__(self, data):
            self.scalar_data = data

    def __init__(self, i_d, data):
        super().__init__(i_d, [ScalarFieldData.ScalarData(_data) for _data in data])


class Vector:
    """Stores the data as a vector ``(x, y, z)``."""

    def __init__(self, x, y, z):
        self._x = x
        self._y = y
        self._z = z

    @property
    def x(self) -> float:
        return self._x

    @property
    def y(self) -> float:
        return self._y

    @property
    def z(self) -> float:
        return self._z


def _resolve_into_array_of_vectors(data):
    if data.size % 3:
        raise ValueError(
            "Dataset must be resolved as a set of vectors."
            "The length of the dataset should always be in multiples of 3."
        )
    data.shape = data.size // 3, 3


class VectorFieldData(BaseFieldData):
    """Provides a container for vector field data."""

    class VectorData(Vector):
        """Stores and provides the data as a vector."""

        def __init__(self, x, y, z):
            super().__init__(x, y, z)

    def __init__(self, i_d, data, scale):
        _resolve_into_array_of_vectors(data)
        self._scale = scale
        super().__init__(i_d, [VectorFieldData.VectorData(x, y, z) for x, y, z in data])

    @property
    def scale(self) -> float:
        return self._scale


class Vertices(BaseFieldData):
    """Provides a container for the vertex data."""

    class Vertex(Vector):
        """Stores and provides the data as a vector of a vertex."""

        def __init__(self, x, y, z):
            super().__init__(x, y, z)

    def __init__(self, i_d, data):
        _resolve_into_array_of_vectors(data)
        super().__init__(i_d, [(Vertices.Vertex(x, y, z)) for x, y, z in data])


class FacesCentroid(BaseFieldData):
    """Provides the container for the face centroid data."""

    class Centroid(Vector):
        """Stores and provides the face centroid data as a vector."""

        def __init__(self, x, y, z):
            super().__init__(x, y, z)

    def __init__(self, i_d, data):
        _resolve_into_array_of_vectors(data)
        super().__init__(i_d, [(FacesCentroid.Centroid(x, y, z)) for x, y, z in data])


class FacesConnectivity(BaseFieldData):
    """Provides the container for the face connectivity data."""

    class Faces:
        """Stores and provides the face connectivity data as an array."""

        def __init__(self, node_count, node_indices):
            self.node_count = node_count
            self.node_indices = node_indices

    def __init__(self, i_d, data):
        faces_data = []
        i = 0

        while i < len(data):
            end = i + 1 + data[i]
            faces_data.append(FacesConnectivity.Faces(data[i], data[i + 1 : end]))
            i = end

        super().__init__(i_d, faces_data)


class FacesNormal(BaseFieldData):
    """Provides the container for the face normal data."""

    class Normal(Vector):
        """Stores and provides the face normal data as a vector."""

        def __init__(self, x, y, z):
            super().__init__(x, y, z)

    def __init__(self, i_d, data):
        _resolve_into_array_of_vectors(data)
        super().__init__(i_d, [FacesNormal.Normal(x, y, z) for x, y, z in data])


class FieldData:
    """Provides access to Fluent field data on surfaces."""

    def __init__(
        self,
        service: FieldDataService,
        field_info: FieldInfo,
        is_data_valid: Callable[[], bool],
    ):
        """__init__ method of FieldData class."""
        self._service = service
        self._field_info = field_info
        self.is_data_valid = is_data_valid

        self._allowed_surface_names = _AllowedSurfaceNames(field_info)

        self._allowed_surface_ids = _AllowedSurfaceIDs(field_info)

        self._allowed_scalar_field_names = _AllowedScalarFieldNames(
            field_info, is_data_valid
        )

        self._allowed_vector_field_names = _AllowedVectorFieldNames(
            field_info, is_data_valid
        )

        surface_args = dict(
            surface_ids=self._allowed_surface_ids,
            surface_name=self._allowed_surface_names,
        )
        scalar_field_args = {
            **dict(field_name=self._allowed_scalar_field_names),
            **surface_args,
        }
        self.get_scalar_field_data = override_help_text(
            _FieldMethod(
                field_data_accessor=self.get_scalar_field_data,
                args_allowed_values_accessors=scalar_field_args,
            ),
            self.get_scalar_field_data,
        )
        self.get_vector_field_data = override_help_text(
            _FieldMethod(
                field_data_accessor=self.get_vector_field_data,
                args_allowed_values_accessors={
                    **dict(field_name=self._allowed_vector_field_names),
                    **surface_args,
                },
            ),
            self.get_vector_field_data,
        )
        self.get_surface_data = override_help_text(
            _FieldMethod(
                field_data_accessor=self.get_surface_data,
                args_allowed_values_accessors=surface_args,
            ),
            self.get_surface_data,
        )
        self.get_pathlines_field_data = override_help_text(
            _FieldMethod(
                field_data_accessor=self.get_pathlines_field_data,
                args_allowed_values_accessors=scalar_field_args,
            ),
            self.get_pathlines_field_data,
        )

    def new_transaction(self):
        """Create a new field transaction."""
        return FieldTransaction(
            self._service,
            self._field_info,
            self._allowed_surface_ids,
            self._allowed_surface_names,
            self._allowed_scalar_field_names,
            self._allowed_vector_field_names,
        )

    def get_scalar_field_data(
        self,
        field_name: str,
        surface_ids: Optional[List[int]] = None,
        surface_name: Optional[str] = None,
        node_value: Optional[bool] = True,
        boundary_value: Optional[bool] = False,
    ) -> Union[ScalarFieldData, Dict[int, ScalarFieldData]]:
        """Get scalar field data on a surface.

        Parameters
        ----------
        field_name : str
            Name of the scalar field.
        surface_ids : List[int], optional
            List of surface IDs for scalar field data.
        surface_name: str, optional
            Surface Name for scalar field data.
        node_value : bool, optional
            Whether to provide data for the nodal location. The default is ``True``.
            When ``False``, data is provided for the element location.
        boundary_value : bool, optional
            Whether to provide slip velocity at the wall boundaries. The default is
            ``False``. When ``True``, no slip velocity is provided.

        Returns
        -------
        Union[ScalarFieldData, Dict[int, ScalarFieldData]]
            If a surface name is provided as input, scalar field data is returned. If surface
            IDs are provided as input, a dictionary containing a map of surface IDs to scalar
            field data.
        """
        surface_ids = _get_surface_ids(
            field_info=self._field_info,
            allowed_surface_names=self._allowed_surface_names,
            surface_ids=surface_ids,
            surface_name=surface_name,
        )
        fields_request = get_fields_request()
        fields_request.scalarFieldRequest.extend(
            [
                FieldDataProtoModule.ScalarFieldRequest(
                    surfaceId=surface_id,
                    scalarFieldName=self._allowed_scalar_field_names.valid_name(
                        field_name
                    ),
                    dataLocation=FieldDataProtoModule.DataLocation.Nodes
                    if node_value
                    else FieldDataProtoModule.DataLocation.Elements,
                    provideBoundaryValues=boundary_value,
                )
                for surface_id in surface_ids
            ]
        )

        fields = extract_fields(self._service.get_fields(fields_request))
        scalar_field_data = next(iter(fields.values()))

        if surface_name:
            return ScalarFieldData(
                surface_ids[0], scalar_field_data[surface_ids[0]][field_name]
            )
        else:
            return {
                surface_id: ScalarFieldData(
                    surface_id, scalar_field_data[surface_id][field_name]
                )
                for surface_id in surface_ids
            }

    def get_surface_data(
        self,
        data_type: SurfaceDataType,
        surface_ids: Optional[List[int]] = None,
        surface_name: Optional[str] = None,
        overset_mesh: Optional[bool] = False,
    ) -> Union[
        Union[Vertices, FacesConnectivity, FacesNormal, FacesCentroid],
        Dict[int, Union[Vertices, FacesConnectivity, FacesNormal, FacesCentroid]],
    ]:
        """Get surface data (vertices, faces connectivity, centroids, and
        normals).

        Parameters
        ----------
        data_type : SurfaceDataType
            SurfaceDataType Enum member.
        surface_ids : List[int], optional
            List of surface IDs for the surface data.
        surface_name : str, optional
            Surface name for the surface data.
        overset_mesh : bool, optional
            Whether to provide the overset method. The default is ``False``.

        Returns
        -------
        Union[Vertices, FacesConnectivity, FacesNormal, FacesCentroid,
        Dict[int, Union[Vertices, FacesConnectivity, FacesNormal, FacesCentroid]]]
             If a surface name is provided as input, face vertices, connectivity data, and normal or centroid data are returned.
             If surface IDs are provided as input, a dictionary containing a map of surface IDs to face
             vertices, connectivity data, and normal or centroid data is returned.
        """
        surface_ids = _get_surface_ids(
            field_info=self._field_info,
            allowed_surface_names=self._allowed_surface_names,
            surface_ids=surface_ids,
            surface_name=surface_name,
        )
        fields_request = get_fields_request()
        fields_request.surfaceRequest.extend(
            [
                FieldDataProtoModule.SurfaceRequest(
                    surfaceId=surface_id,
                    oversetMesh=overset_mesh,
                    provideFaces=data_type == SurfaceDataType.FacesConnectivity,
                    provideVertices=data_type == SurfaceDataType.Vertices,
                    provideFacesCentroid=data_type == SurfaceDataType.FacesCentroid,
                    provideFacesNormal=data_type == SurfaceDataType.FacesNormal,
                )
                for surface_id in surface_ids
            ]
        )
        enum_to_field_name = {
            SurfaceDataType.FacesConnectivity: "faces",
            SurfaceDataType.Vertices: "vertices",
            SurfaceDataType.FacesCentroid: "centroid",
            SurfaceDataType.FacesNormal: "face-normal",
        }
        fields = extract_fields(self._service.get_fields(fields_request))
        surface_data = next(iter(fields.values()))

        def _get_surfaces_data(parent_class, surf_id, _data_type):
            return parent_class(
                surf_id,
                surface_data[surf_id][enum_to_field_name[_data_type]],
            )

        if data_type == SurfaceDataType.Vertices:
            if surface_name:
                return _get_surfaces_data(Vertices, surface_ids[0], data_type)
            else:
                return {
                    surface_id: _get_surfaces_data(Vertices, surface_id, data_type)
                    for surface_id in surface_ids
                }

        if data_type == SurfaceDataType.FacesCentroid:
            if surface_name:
                return _get_surfaces_data(FacesCentroid, surface_ids[0], data_type)
            else:
                return {
                    surface_id: _get_surfaces_data(FacesCentroid, surface_id, data_type)
                    for surface_id in surface_ids
                }

        if data_type == SurfaceDataType.FacesConnectivity:
            if surface_name:
                return _get_surfaces_data(FacesConnectivity, surface_ids[0], data_type)
            else:
                return {
                    surface_id: _get_surfaces_data(
                        FacesConnectivity, surface_id, data_type
                    )
                    for surface_id in surface_ids
                }

        if data_type == SurfaceDataType.FacesNormal:
            if surface_name:
                return _get_surfaces_data(FacesNormal, surface_ids[0], data_type)
            else:
                return {
                    surface_id: _get_surfaces_data(FacesNormal, surface_id, data_type)
                    for surface_id in surface_ids
                }

    def get_vector_field_data(
        self,
        field_name: str,
        surface_ids: Optional[List[int]] = None,
        surface_name: Optional[str] = None,
    ) -> Union[VectorFieldData, Dict[int, VectorFieldData]]:
        """Get vector field data on a surface.

        Parameters
        ----------
        field_name : str
            Name of the vector field.
        surface_ids : List[int], optional
            List of surface IDs for vector field data.
        surface_name: str, optional
            Surface Name for vector field data.

        Returns
        -------
        Union[VectorFieldData, Dict[int, VectorFieldData]]
            If a surface name is provided as input, vector field data is returned.
            If surface IDs are provided as input, a dictionary containing a map of
            surface IDs to vector field data is returned.
        """
        surface_ids = _get_surface_ids(
            field_info=self._field_info,
            allowed_surface_names=self._allowed_surface_names,
            surface_ids=surface_ids,
            surface_name=surface_name,
        )
        fields_request = get_fields_request()
        fields_request.vectorFieldRequest.extend(
            [
                FieldDataProtoModule.VectorFieldRequest(
                    surfaceId=surface_id,
                    vectorFieldName=self._allowed_vector_field_names.valid_name(
                        field_name
                    ),
                )
                for surface_id in surface_ids
            ]
        )
        fields = extract_fields(self._service.get_fields(fields_request))
        vector_field_data = next(iter(fields.values()))

        if surface_name:
            return VectorFieldData(
                surface_ids[0],
                vector_field_data[surface_ids[0]][field_name],
                vector_field_data[surface_ids[0]]["vector-scale"][0],
            )
        else:
            return {
                surface_id: VectorFieldData(
                    surface_id,
                    vector_field_data[surface_id][field_name],
                    vector_field_data[surface_id]["vector-scale"][0],
                )
                for surface_id in surface_ids
            }

    def get_pathlines_field_data(
        self,
        field_name: str,
        surface_ids: Optional[List[int]] = None,
        surface_name: Optional[str] = None,
        additional_field_name: Optional[str] = "",
        provide_particle_time_field: Optional[bool] = False,
        node_value: Optional[bool] = True,
        steps: Optional[int] = 500,
        step_size: Optional[float] = 0.01,
        skip: Optional[int] = 0,
        reverse: Optional[bool] = False,
        accuracy_control_on: Optional[bool] = False,
        tolerance: Optional[float] = 0.001,
        coarsen: Optional[int] = 1,
        velocity_domain: Optional[str] = "all-phases",
        zones: Optional[list] = [],
    ) -> Dict:
        """Get the pathlines field data on a surface.

        Parameters
        ----------
        field_name : str
            Name of the scalar field to color pathlines.
        surface_ids : List[int], optional
            List of surface IDs for pathlines field data.
        surface_name : str, optional
            Surface name for pathlines field data.
        additional_field_name : str, optional
            Additional field if required.
        provide_particle_time_field: bool, optional
            Whether to provide the particle time. The default is ``False``.
        node_value : bool, optional
                    Whether to provide the nodal values. The default is ``True``. If
                    ``False``, element values are provided.
        steps: int, optional
            Pathlines steps. The default is ``500``
        step_size: float, optional
            Pathlines step size. The default is ``0.01``.
        skip: int, optional
            Pathlines to skip. The default is ``0``.
        reverse: bool, optional
            Whether to draw pathlines in reverse direction. The default is ``False``.
        tolerance: float, optional
            Pathlines tolerance. The default is ``0.001``.
        coarsen: int, optional
            Pathlines coarsen. The default is ``1``.
        velocity_domain: str, optional
            Domain for pathlines. The default is ``"all-phases"``.
        zones: list, optional
            Zones for pathlines. The default is ``[]``.

        Returns
        -------
        Dict
            Dictionary containing a map of surface IDs to the pathline data.
            For example, pathlines connectivity, vertices, and field.
        """
        surface_ids = _get_surface_ids(
            field_info=self._field_info,
            allowed_surface_names=self._allowed_surface_names,
            surface_ids=surface_ids,
            surface_name=surface_name,
        )
        fields_request = get_fields_request()
        fields_request.pathlinesFieldRequest.extend(
            [
                FieldDataProtoModule.PathlinesFieldRequest(
                    surfaceId=surface_id,
                    field=field_name,
                    additionalField=additional_field_name,
                    provideParticleTimeField=provide_particle_time_field,
                    dataLocation=FieldDataProtoModule.DataLocation.Nodes
                    if node_value
                    else FieldDataProtoModule.DataLocation.Elements,
                    steps=steps,
                    stepSize=step_size,
                    skip=skip,
                    reverse=reverse,
                    accuracyControlOn=accuracy_control_on,
                    tolerance=tolerance,
                    coarsen=coarsen,
                    velocityDomain=velocity_domain,
                    zones=zones,
                )
                for surface_id in surface_ids
            ]
        )
        fields = extract_fields(self._service.get_fields(fields_request))
        pathlines_data = next(iter(fields.values()))

        def _get_surfaces_data(parent_class, surf_id, _data_type):
            return parent_class(
                surf_id,
                pathlines_data[surf_id][_data_type],
            )

        if surface_name:
            vertices_data = _get_surfaces_data(Vertices, surface_ids[0], "vertices")
            lines_data = _get_surfaces_data(FacesConnectivity, surface_ids[0], "lines")
            field_data = ScalarFieldData(
                surface_ids[0], pathlines_data[surface_ids[0]][field_name]
            )
            return {
                "vertices": vertices_data,
                "lines": lines_data,
                field_name: field_data,
            }
        else:
            path_lines_dict = {}
            for surface_id in surface_ids:
                path_lines_dict[surface_id] = {
                    "vertices": _get_surfaces_data(Vertices, surface_id, "vertices"),
                    "lines": _get_surfaces_data(FacesConnectivity, surface_id, "lines"),
                    field_name: ScalarFieldData(
                        surface_id, pathlines_data[surface_id][field_name]
                    ),
                }
            return path_lines_dict



"""Module providing visualization objects to facilitate
   integration with libraries like Matplotlib and pyvista."""
import inspect

from ansys.fluent.core.meta import PyLocalContainer


class Container:
    """
    Base class for containers, e.g. Plots, Graphics.

    Parameters
        ----------
        session : object
            Session object.
        container_type: object
            Container type (e.g. Plots, Graphics)
        module: object
            Python module containing post definitions
        post_api_helper: object
            Provides helper APIs for post-processing
        local_surfaces_provider : object, optional
            Object providing local surfaces so that user can access surfaces
            created in other modules, such as PyVista. The default is ``None``.
    """

    def __init__(
        self,
        session,
        container_type,
        module,
        post_api_helper,
        local_surfaces_provider=None,
    ):
        session_state = container_type._sessions_state.get(session)
        if not session_state:
            session_state = self.__dict__
            container_type._sessions_state[session] = session_state
            self.session = session
            self._init_module(self, module, post_api_helper)
        else:
            self.__dict__ = session_state
        self._local_surfaces_provider = lambda: local_surfaces_provider or getattr(
            self, "Surfaces", []
        )

    @property
    def type(self):
        return "object"

    def _init_module(self, obj, mod, post_api_helper):
        for name, cls in mod.__dict__.items():
            if cls.__class__.__name__ in (
                "PyLocalNamedObjectMetaAbstract",
            ) and not inspect.isabstract(cls):
                setattr(
                    obj,
                    cls.PLURAL,
                    PyLocalContainer(self, cls, post_api_helper),
                )


class Plots(Container):
    """Provides the Matplotlib ``Plots`` objects manager.

    This class provides access to ``Plots`` object containers for a given
    session so that plots can be created.

    Parameters
        ----------
        session : obj
            Session object.
        module: object
            Python module containing post definitions
        post_api_helper: object
            Provides helper APIs for post-processing
        local_surfaces_provider : object, optional
            Object providing local surfaces so that you can access surfaces
            created in other modules, such as pyvista. The default is ``None``.

    Attributes
    ----------
    XYPlots : dict
        Container for XY plot objects.
    MonitorPlots : dict
        Container for monitor plot objects.
    """

    _sessions_state = {}

    def __init__(self, session, module, post_api_helper, local_surfaces_provider=None):
        super().__init__(
            session, self.__class__, module, post_api_helper, local_surfaces_provider
        )


class Graphics(Container):
    """Provides the pyvista ``Graphics`` objects manager.

    This class provides access to ``Graphics`` object containers for a given
    session so that graphics objects can be created.

    Parameters
    ----------
    session : obj
        Session object.
    module: object
        Python module containing post definitions
    post_api_helper: object
        Provides helper APIs for post-processing
    local_surfaces_provider : object, optional
        Object providing local surfaces so that you can access surfaces
        created in other modules, such as pyvista. The default is ``None``.

    Attributes
    ----------
    Meshes : dict
        Container for mesh objects.
    Surfaces : dict
        Container for surface objects.
    Contours : dict
        Container for contour objects.
    Vectors : dict
        Container for vector objects.
    """

    _sessions_state = {}

    def __init__(self, session, module, post_api_helper, local_surfaces_provider=None):
        super().__init__(
            session, self.__class__, module, post_api_helper, local_surfaces_provider
        )

    def add_outline_mesh(self):
        """Add a mesh outline.

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        meshes = getattr(self, "Meshes", None)
        if meshes is not None:
            outline_mesh_id = "Mesh-outline"
            outline_mesh = meshes[outline_mesh_id]
            outline_mesh.surfaces_list = [
                k
                for k, v in outline_mesh._api_helper.field_info()
                .get_surfaces_info()
                .items()
                if v["type"] == "zone-surf" and v["zone_type"] != "interior"
            ]
            return outline_mesh



#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import _ChildNamedObjectAccessorMixin

from ansys.fluent.core.solver.flobject import _CreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _NonCreatableNamedObjectMixin

class vapor_phase_realgas(Integer):
    """
    'vapor_phase_realgas' child.
    """

    fluent_name = "vapor-phase-realgas"




"""Module containing class encapsulating Fluent connection.
"""


import functools

from ansys.api.fluent.v0 import datamodel_se_pb2
import ansys.fluent.core as pyfluent
from ansys.fluent.core.data_model_cache import DataModelCache
from ansys.fluent.core.fluent_connection import FluentConnection
from ansys.fluent.core.services.streaming import StreamingService
from ansys.fluent.core.session import BaseSession
from ansys.fluent.core.session_base_meshing import BaseMeshing
from ansys.fluent.core.streaming_services.datamodel_streaming import DatamodelStream
from ansys.fluent.core.utils.data_transfer import transfer_case


class PureMeshing(BaseSession):
    """Encapsulates a Fluent meshing session. A ``tui`` object
    for meshing TUI commanding, and ``meshing`` and ``workflow``
    objects for access to task-based meshing workflows are all
    exposed here. No ``switch_to_solver`` method is available
    in this mode."""

    rules = ["workflow", "meshing", "PartManagement", "PMFileManagement"]
    for r in rules:
        DataModelCache.set_config(r, "internal_names_as_keys", True)

    def __init__(self, fluent_connection: FluentConnection):
        """PureMeshing session.

        Args:
            fluent_connection (:ref:`ref_fluent_connection`): Encapsulates a Fluent connection.
        """
        super(PureMeshing, self).__init__(fluent_connection=fluent_connection)
        self._base_meshing = BaseMeshing(self.execute_tui, fluent_connection)
        datamodel_service_se = fluent_connection.datamodel_service_se
        self.datamodel_streams = {}
        if pyfluent.DATAMODEL_USE_STATE_CACHE:
            for rules in self.__class__.rules:
                request = datamodel_se_pb2.DataModelRequest()
                request.rules = rules
                if pyfluent.DATAMODEL_USE_NOCOMMANDS_DIFF_STATE:
                    request.diffstate = datamodel_se_pb2.DIFFSTATE_NOCOMMANDS
                streaming = StreamingService(
                    stub=datamodel_service_se._stub,
                    request=request,
                    metadata=datamodel_service_se._metadata,
                )
                stream = DatamodelStream(streaming)
                stream.register_callback(
                    functools.partial(DataModelCache.update_cache, rules=rules)
                )
                self.datamodel_streams[rules] = stream
                stream.start()

    @property
    def tui(self):
        """Instance of ``main_menu`` on which Fluent's SolverTUI methods can be
        executed."""
        return self._base_meshing.tui

    @property
    def meshing(self):
        """meshing datamodel root."""
        return self._base_meshing.meshing

    @property
    def workflow(self):
        """workflow datamodel root."""
        return self._base_meshing.workflow

    @property
    def PartManagement(self):
        """PartManagement datamodel root."""
        return self._base_meshing.PartManagement

    @property
    def PMFileManagement(self):
        """PMFileManagement datamodel root."""
        return self._base_meshing.PMFileManagement

    @property
    def preferences(self):
        """preferences datamodel root."""
        return self._base_meshing.preferences

    def transfer_mesh_to_solvers(
        self,
        solvers,
        file_type: str = "case",
        file_name_stem: str = None,
        num_files_to_try: int = 1,
        clean_up_mesh_file: bool = True,
        overwrite_previous: bool = True,
    ):
        """Transfer mesh to Fluent solver instances.

        Parameters
        ----------
        solvers : iterable
            Sequence of solver instances
        file_type : str, default "case"
            "case" or "mesh"
        file_name_stem : str
            Optional file name stem
        num_files_to_try : int, default 1
            Optional number of files to try to write,
            each with a different generated name.
            Defaults to 1
        clean_up_mesh_file: bool, default True
            Whether to remove the file at the end
        overwrite_previous: bool, default True
            Whether to overwrite the file if it already exists
        Returns
        -------
        None
        """
        transfer_case(
            self,
            solvers,
            file_type,
            file_name_stem,
            num_files_to_try,
            clean_up_mesh_file,
            overwrite_previous,
        )



#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import _ChildNamedObjectAccessorMixin

from ansys.fluent.core.solver.flobject import _CreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _NonCreatableNamedObjectMixin

from .option_9 import option
from .value import value
class premix_heat_trans_coeff(Group):
    """
    'premix_heat_trans_coeff' child.
    """

    fluent_name = "premix-heat-trans-coeff"

    child_names = \
        ['option', 'value']

    option: option = option
    """
    option child of premix_heat_trans_coeff.
    """
    value: value = value
    """
    value child of premix_heat_trans_coeff.
    """



#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import _ChildNamedObjectAccessorMixin

from ansys.fluent.core.solver.flobject import _CreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _NonCreatableNamedObjectMixin

class user_defined_value(Boolean):
    """
    'user_defined_value' child.
    """

    fluent_name = "user-defined-value?"




#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import _ChildNamedObjectAccessorMixin

from ansys.fluent.core.solver.flobject import _CreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _NonCreatableNamedObjectMixin

class wall_omega_treatment(String):
    """
    Specify a wall omega treatment.
    """

    fluent_name = "wall-omega-treatment"




#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import _ChildNamedObjectAccessorMixin

from ansys.fluent.core.solver.flobject import _CreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _NonCreatableNamedObjectMixin

class relaxation_options(String):
    """
    Enter high order relaxation option with respect to diffusion gradient.
    """

    fluent_name = "relaxation-options"




#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import _ChildNamedObjectAccessorMixin

from ansys.fluent.core.solver.flobject import _CreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _NonCreatableNamedObjectMixin

from .nb_gradient import nb_gradient
from .nb_gradient_dbns import nb_gradient_dbns
class nb_gradient_boundary_option(Group):
    """
    'nb_gradient_boundary_option' child.
    """

    fluent_name = "nb-gradient-boundary-option?"

    child_names = \
        ['nb_gradient', 'nb_gradient_dbns']

    nb_gradient: nb_gradient = nb_gradient
    """
    nb_gradient child of nb_gradient_boundary_option.
    """
    nb_gradient_dbns: nb_gradient_dbns = nb_gradient_dbns
    """
    nb_gradient_dbns child of nb_gradient_boundary_option.
    """



#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import _ChildNamedObjectAccessorMixin

from ansys.fluent.core.solver.flobject import _CreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _NonCreatableNamedObjectMixin

from .report_plots_child import report_plots_child

class report_plots(NamedObject[report_plots_child], _CreatableNamedObjectMixin[report_plots_child]):
    """
    'report_plots' child.
    """

    fluent_name = "report-plots"

    child_object_type: report_plots_child = report_plots_child
    """
    child_object_type of report_plots.
    """



"""Client-side service allowing access and modification of rpvars.

The primary interaction with Fluent should not be through low-level
variables like rpvars but instead through the high-level object-based
interfaces: solver settings objects and task-based meshing workflow.
"""
from typing import Any, List

import ansys.fluent.core.filereader.lispy as lispy
from ansys.fluent.core.solver.error_message import allowed_name_error_message


class RPVars:
    """Access to rpvars in a specific session.

    Methods
    -------
    __call__(var, val)
        Set or get a specific rpvar or get the full rpvar state.
    """

    _allowed_values = None

    def __init__(self, eval_fn):
        self._eval_fn = eval_fn

    def __call__(self, var: str = None, val: Any = None) -> Any:
        """Set or get a specific rpvar, or get the full rpvar state.

        Parameters
        ----------
        var : str
            Name of the rpvar. Set the specified rpvar if val is provided,
            or else return its value.
        val : Any
            Value of the rpvar. Set the specified var to this value if val
            is specified.

        Returns
        -------
        Any
            A dict containing the full rpvar state if all arguments are
            unspecified, or the value of the rpvar if only var is specified,
            or None if both arguments are specified.
        """
        return (
            self._set_var(var, val)
            if val
            else (self._get_var(var) if var else self._get_vars())
        )

    def allowed_values(self) -> List[str]:
        if not RPVars._allowed_values:
            RPVars._allowed_values = lispy.parse(
                self._eval_fn("(cx-send '(map car rp-variables))")
            )
        return RPVars._allowed_values

    def _get_var(self, var: str):
        if var not in self.allowed_values():
            raise RuntimeError(
                allowed_name_error_message("rp-vars", var, RPVars._allowed_values)
            )

        cmd = f"(rpgetvar {RPVars._var(var)})"
        return self._execute(cmd)

    def _get_vars(self):
        list_val = self._execute("(cx-send 'rp-variables)")
        return {val[0]: val[1] for val in list_val}

    def _set_var(self, var: str, val):
        prefix = "'" if isinstance(val, (list, tuple)) else ""
        cmd = f"(rpsetvar {RPVars._var(var)} {prefix}{lispy.to_string(val)})"
        return self._execute(cmd)

    def _execute(self, cmd: str):
        scheme_val = self._eval_fn(cmd)
        return lispy.parse(scheme_val)

    @staticmethod
    def _var(var: str):
        if not var.startswith("'"):
            var = "'" + var
        return var



import pytest
from util.solver_workflow import new_solver_session  # noqa: F401

from ansys.fluent.core import examples


@pytest.mark.fluent_232
def test_1364(new_solver_session):
    solver = new_solver_session

    import_filename = examples.download_file(
        "elbow.cas.h5", "pyfluent/examples/DOE-ML-Mixing-Elbow"
    )

    solver.file.read_case(file_name=import_filename)

    solver.solution.report_definitions.volume.create("xxx")

    solver.solution.report_definitions["xxx"].set_state(
        {
            "report_type": "volume-max",
            "field": "temperature",
            "average_over": 1,
            "per_zone": False,
            "zone_names": ["fluid"],
            "expr_list": None,
        }
    )

    assert solver.solution.report_definitions.volume[
        "xxx"
    ].zone_names.allowed_values() == ["fluid"]

    assert (
        solver.solution.report_definitions.volume["xxx"].expr_list.allowed_values()
        == None
    )



#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import _ChildNamedObjectAccessorMixin

from ansys.fluent.core.solver.flobject import _CreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _NonCreatableNamedObjectMixin

from .pre_sweeps_2 import pre_sweeps
from .post_sweeps_2 import post_sweeps
class fixed_cycle_parameters(Group):
    """
    'fixed_cycle_parameters' child.
    """

    fluent_name = "fixed-cycle-parameters"

    child_names = \
        ['pre_sweeps', 'post_sweeps']

    pre_sweeps: pre_sweeps = pre_sweeps
    """
    pre_sweeps child of fixed_cycle_parameters.
    """
    post_sweeps: post_sweeps = post_sweeps
    """
    post_sweeps child of fixed_cycle_parameters.
    """



import numpy as np
import pytest
from util.solver_workflow import new_solver_session  # noqa: F401

from ansys.fluent.core import examples
from ansys.fluent.core.services.field_data import (
    ScalarFieldNameError,
    ScalarFieldUnavailable,
    SurfaceDataType,
    SurfaceNameError,
)

HOT_INLET_TEMPERATURE = 313.15


@pytest.mark.dev
@pytest.mark.fluent_232
def test_field_data(new_solver_session) -> None:
    solver = new_solver_session
    import_filename = examples.download_file(
        "mixing_elbow.msh.h5", "pyfluent/mixing_elbow"
    )
    solver.file.read(file_type="case", file_name=import_filename)
    solver.tui.mesh.check()

    solver.setup.models.energy.enabled = True
    solver.setup.materials.database.copy_by_name(type="fluid", name="water-liquid")
    solver.setup.cell_zone_conditions.fluid["elbow-fluid"].material = "water-liquid"

    # Set up boundary conditions for CFD analysis
    solver.setup.boundary_conditions.velocity_inlet["cold-inlet"].vmag = 0.4
    solver.setup.boundary_conditions.velocity_inlet[
        "cold-inlet"
    ].ke_spec = "Intensity and Hydraulic Diameter"
    solver.setup.boundary_conditions.velocity_inlet["cold-inlet"].turb_intensity = 0.05
    solver.setup.boundary_conditions.velocity_inlet[
        "cold-inlet"
    ].turb_hydraulic_diam = "4 [in]"

    solver.setup.boundary_conditions.velocity_inlet["cold-inlet"].t = 293.15

    solver.setup.boundary_conditions.velocity_inlet["hot-inlet"].vmag = 1.2
    solver.setup.boundary_conditions.velocity_inlet[
        "hot-inlet"
    ].ke_spec = "Intensity and Hydraulic Diameter"
    solver.setup.boundary_conditions.velocity_inlet[
        "hot-inlet"
    ].turb_hydraulic_diam = "1 [in]"

    solver.setup.boundary_conditions.velocity_inlet[
        "hot-inlet"
    ].t = HOT_INLET_TEMPERATURE

    solver.setup.boundary_conditions.pressure_outlet["outlet"].turb_viscosity_ratio = 4

    solver.tui.solve.monitors.residual.plot("no")

    # Initialize flow field
    solver.solution.initialization.hybrid_initialize()

    solver.solution.run_calculation.iterate.get_attr("arguments")
    solver.solution.run_calculation.iterate(iter_count=10)

    # Get field data object
    field_data = solver.field_data

    transaction = field_data.new_transaction()

    hot_inlet_surf_id = solver.field_info.get_surfaces_info()["hot-inlet"][
        "surface_id"
    ][0]
    transaction.add_surfaces_request(
        surface_ids=[1, hot_inlet_surf_id],
        provide_vertices=True,
        provide_faces=False,
        provide_faces_centroid=True,
    )
    transaction.add_scalar_fields_request(
        surface_ids=[1, hot_inlet_surf_id],
        field_name="temperature",
        node_value=True,
        boundary_value=True,
    )
    transaction.add_pathlines_fields_request(
        surface_ids=[1, hot_inlet_surf_id],
        field_name="temperature",
        provide_particle_time_field=True,
    )

    data = transaction.get_fields()

    surface_data_tag = (("type", "surface-data"),)  # tuple containing surface data info
    scalar_field_tag = (
        ("type", "scalar-field"),
        ("dataLocation", 0),
        ("boundaryValues", True),
    )  # tuple containing scalar field info
    pathline_tag = (("type", "pathlines-field"), ("field", "temperature"))
    assert len(data) == 3
    assert list(data[surface_data_tag][hot_inlet_surf_id].keys()) == [
        "vertices",
        "centroid",
    ]
    assert list(data[scalar_field_tag][hot_inlet_surf_id].keys()) == ["temperature"]
    assert (
        len(data[scalar_field_tag][hot_inlet_surf_id]["temperature"])
        == len(data[surface_data_tag][hot_inlet_surf_id]["vertices"]) / 3
    )
    assert (
        round(
            float(np.average(data[scalar_field_tag][hot_inlet_surf_id]["temperature"])),
            2,
        )
        == HOT_INLET_TEMPERATURE
    )
    assert sorted(list(data[pathline_tag][hot_inlet_surf_id].keys())) == sorted(
        [
            "vertices",
            "lines",
            "temperature",
            "pathlines-count",
            "particle-time",
        ]
    )

    # multiple surface *names* transaction
    transaction2 = field_data.new_transaction()
    surface_names = (
        transaction2.add_scalar_fields_request.surface_names.allowed_values()
    )
    transaction2.add_scalar_fields_request(
        surface_names=surface_names, field_name="temperature"
    )
    data2 = transaction2.get_fields()
    assert data2


def test_field_data_allowed_values(new_solver_session) -> None:
    solver = new_solver_session
    import_filename = examples.download_file(
        "mixing_elbow.msh.h5", "pyfluent/mixing_elbow"
    )

    field_data = solver.field_data
    field_info = solver.field_info
    transaction = field_data.new_transaction()

    assert [] == field_data.get_scalar_field_data.field_name.allowed_values()

    solver.file.read(file_type="case", file_name=import_filename)

    allowed_args_no_init = field_data.get_scalar_field_data.field_name.allowed_values()
    assert len(allowed_args_no_init) != 0

    assert not field_data.is_data_valid()

    solver.solution.initialization.hybrid_initialize()

    assert field_data.is_data_valid()

    expected_allowed_args = sorted(field_info.get_scalar_fields_info())
    allowed_args = field_data.get_scalar_field_data.field_name.allowed_values()
    assert expected_allowed_args and (expected_allowed_args == allowed_args)
    assert len(allowed_args) > len(allowed_args_no_init)
    allowed_args = transaction.add_scalar_fields_request.field_name.allowed_values()
    assert expected_allowed_args == allowed_args

    expected_allowed_args = sorted(field_info.get_surfaces_info())
    allowed_args = field_data.get_scalar_field_data.surface_name.allowed_values()
    assert expected_allowed_args and (expected_allowed_args == allowed_args)
    allowed_args = transaction.add_scalar_fields_request.surface_names.allowed_values()
    assert expected_allowed_args == allowed_args

    expected_allowed_args = sorted(field_info.get_surfaces_info())
    allowed_args = field_data.get_surface_data.surface_name.allowed_values()
    assert expected_allowed_args and (expected_allowed_args == allowed_args)
    allowed_args = transaction.add_scalar_fields_request.surface_names.allowed_values()
    assert expected_allowed_args == allowed_args

    allowed_args = field_data.get_surface_data.surface_ids.allowed_values()
    assert len(expected_allowed_args) == len(allowed_args)
    allowed_args = transaction.add_scalar_fields_request.surface_ids.allowed_values()
    assert len(expected_allowed_args) == len(allowed_args)

    expected_allowed_args = sorted(field_info.get_vector_fields_info())
    allowed_args = field_data.get_vector_field_data.field_name.allowed_values()
    assert expected_allowed_args and (expected_allowed_args == allowed_args)
    allowed_args = transaction.add_vector_fields_request.field_name.allowed_values()
    assert expected_allowed_args == allowed_args


@pytest.mark.dev
@pytest.mark.fluent_232
def test_field_data_objects_3d(new_solver_session) -> None:
    solver = new_solver_session
    import_filename = examples.download_file(
        "mixing_elbow.msh.h5", "pyfluent/mixing_elbow"
    )

    field_data = solver.field_data

    assert [] == field_data.get_scalar_field_data.field_name.allowed_values()

    solver.file.read(file_type="case", file_name=import_filename)

    allowed_args_no_init = field_data.get_scalar_field_data.field_name.allowed_values()
    assert len(allowed_args_no_init) != 0

    assert not field_data.is_data_valid()

    solver.solution.initialization.hybrid_initialize()

    assert field_data.is_data_valid()

    # Absolute Pressure data over the cold-inlet (surface_id=3)
    abs_press_data = field_data.get_scalar_field_data(
        field_name="absolute-pressure", surface_name="cold-inlet"
    )

    assert abs_press_data.size == 241
    assert abs_press_data[120].scalar_data == 101325.0

    vertices_data = field_data.get_surface_data(
        data_type=SurfaceDataType.Vertices, surface_name="cold-inlet"
    )
    assert round(float(vertices_data[5].x), 2) == -0.2

    faces_centroid_data = field_data.get_surface_data(
        data_type=SurfaceDataType.FacesCentroid, surface_name="cold-inlet"
    )
    assert round(float(faces_centroid_data[5].y), 2) == -0.18

    faces_connectivity_data = field_data.get_surface_data(
        data_type=SurfaceDataType.FacesConnectivity, surface_name="cold-inlet"
    )
    assert faces_connectivity_data[5].node_count == 4
    assert (faces_connectivity_data[5].node_indices == [12, 13, 17, 16]).all()

    faces_normal_data = field_data.get_surface_data(
        data_type=SurfaceDataType.FacesNormal, surface_name="cold-inlet"
    )
    assert faces_normal_data.size == 152
    assert faces_normal_data.surface_id == 3

    velocity_vector_data = field_data.get_vector_field_data(
        field_name="velocity", surface_name="cold-inlet"
    )

    assert velocity_vector_data.size == 152
    assert velocity_vector_data.scale == 1.0

    path_lines_data = field_data.get_pathlines_field_data(
        field_name="velocity", surface_name="cold-inlet"
    )

    assert path_lines_data["vertices"].size == 76152
    assert path_lines_data["lines"].size == 76000
    assert path_lines_data["velocity"].size == 76152

    assert path_lines_data["lines"][100].node_count == 2
    assert all(path_lines_data["lines"][100].node_indices == [100, 101])


@pytest.mark.dev
@pytest.mark.fluent_232
def test_field_data_objects_2d(load_disk_mesh) -> None:
    solver = load_disk_mesh

    field_data = solver.field_data

    allowed_args_no_init = field_data.get_scalar_field_data.field_name.allowed_values()
    assert len(allowed_args_no_init) != 0

    assert not field_data.is_data_valid()

    solver.solution.initialization.hybrid_initialize()

    assert field_data.is_data_valid()

    # Absolute Pressure data over the cold-inlet (surface_id=3)
    abs_press_data = field_data.get_scalar_field_data(
        field_name="absolute-pressure", surface_name="velocity-inlet-2"
    )

    assert abs_press_data.size == 11
    assert abs_press_data[5].scalar_data == 101325.0

    vertices_data = field_data.get_surface_data(
        data_type=SurfaceDataType.Vertices, surface_name="interior-4"
    )
    assert round(float(vertices_data[5].x), 2) == 0.0

    faces_centroid_data = field_data.get_surface_data(
        data_type=SurfaceDataType.FacesCentroid, surface_name="velocity-inlet-2"
    )
    assert round(float(faces_centroid_data[5].y), 2) == 0.02

    faces_connectivity_data = field_data.get_surface_data(
        data_type=SurfaceDataType.FacesConnectivity, surface_name="velocity-inlet-2"
    )
    assert faces_connectivity_data[5].node_count == 2
    assert (faces_connectivity_data[5].node_indices == [5, 6]).all()

    velocity_vector_data = field_data.get_vector_field_data(
        field_name="velocity", surface_name="velocity-inlet-2"
    )

    assert velocity_vector_data.size == 10
    assert velocity_vector_data.scale == 1.0

    path_lines_data = field_data.get_pathlines_field_data(
        field_name="velocity", surface_name="velocity-inlet-2"
    )

    assert path_lines_data["vertices"].size == 5010
    assert path_lines_data["lines"].size == 5000
    assert path_lines_data["velocity"].size == 5010

    assert path_lines_data["lines"][100].node_count == 2
    assert all(path_lines_data["lines"][100].node_indices == [100, 101])


def test_field_data_errors(new_solver_session) -> None:
    solver = new_solver_session
    import_filename = examples.download_file(
        "mixing_elbow.msh.h5", "pyfluent/mixing_elbow"
    )

    with pytest.raises(ScalarFieldNameError) as fne:
        solver.field_data.get_scalar_field_data(
            field_name="y-face-area", surface_ids=[0]
        )
    assert fne.value.field_name == "y-face-area"

    with pytest.raises(ScalarFieldNameError) as fne:
        solver.field_data.get_scalar_field_data(
            field_name="partition-neighbors", surface_ids=[0]
        )
    assert fne.value.field_name == "partition-neighbors"

    solver.file.read(file_type="case", file_name=import_filename)

    with pytest.raises(ScalarFieldUnavailable) as fnu:
        solver.field_data.get_scalar_field_data(field_name="density", surface_ids=[0])
    assert fnu.value.field_name == "density"

    y_face_area = solver.field_data.get_scalar_field_data(
        field_name="y-face-area", surface_ids=[0]
    )
    assert y_face_area and isinstance(y_face_area, dict)

    partition_neighbors = solver.field_data.get_scalar_field_data(
        field_name="partition-neighbors", surface_ids=[0]
    )
    assert partition_neighbors and isinstance(partition_neighbors, dict)

    # Initialize flow field
    solver.solution.initialization.hybrid_initialize()

    # Get field data object
    field_data = solver.field_data

    with pytest.raises(SurfaceNameError) as sne:
        solver.field_data.get_scalar_field_data(
            field_name="density", surface_name="bob"
        )
    assert sne.value.surface_name == "bob"

    with pytest.raises(ScalarFieldNameError) as fne:
        solver.field_data.get_scalar_field_data(field_name="xdensity", surface_ids=[0])
    assert fne.value.field_name == "xdensity"



import math

import pytest

import ansys.fluent.core.quantity as q

DELTA = 1.0e-5


def test_value_unit_1():
    v = q.Quantity(1, "m s^-1")
    assert v.value == 1
    assert v.unit == "m s^-1"


def test_value_unit_2():
    v = q.Quantity(10.6, "m")
    assert v.value == 10.6
    assert v.unit == "m"


def test_value_unit_3():
    v = q.Quantity(99.85, "radian")
    assert v.value == pytest.approx(99.85, DELTA)
    assert v.unit == "radian"


def test_value_unit_4():
    v = q.Quantity(5.7, "")
    assert v.value == pytest.approx(5.7, DELTA)
    assert v.unit == ""


def test_dims_5():
    v = q.Quantity(1.0, "kPa")
    assert v.get_dimensions_list() == [
        1.0,
        -1.0,
        -2.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
    ]


def test_dims_6():
    v = q.Quantity(1.0, "ft")
    assert v.get_dimensions_list() == [0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]


def test_dims_7():
    v = q.Quantity(1.0, "m m^-1")
    assert v.get_dimensions_list() == [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]


def test_to_8():
    v = q.Quantity(1.0, "m")
    convert = v.to("ft")
    assert convert.value == pytest.approx(3.2808398, DELTA)
    assert convert.unit == "ft"


def test_to_9():
    v = q.Quantity(1.0, "m")
    convert = v.to("mm")
    assert convert.value == 1000
    assert convert.unit == "mm"


def test_to_10():
    v = q.Quantity(100000.0, "Pa")
    convert = v.to("kPa")
    assert convert.value == 100.0
    assert convert.unit == "kPa"


def test_to_11():
    v = q.Quantity(1.0, "dm^3")
    convert = v.to("m^3")
    assert convert.value == pytest.approx(0.001, DELTA)
    assert convert.unit == "m^3"


def test_to_12():
    v = q.Quantity(1.0, "radian")
    convert = v.to("degree")
    assert convert.value == pytest.approx(57.295779, DELTA)
    assert convert.unit == "degree"


def test_to_13():
    v = q.Quantity(1.0, "degree")
    convert = v.to("radian")
    assert convert.value == pytest.approx(0.01745329251, DELTA)
    assert convert.unit == "radian"


def test_to_14():
    v = q.Quantity(1.0, "Pa s")
    convert = v.to("dyne cm^-2 s")
    assert convert.value == pytest.approx(10.0, DELTA)
    assert convert.unit == "dyne cm^-2 s"


def test_to_15():
    v = q.Quantity(1.0, "kg m^-1 s^-1")
    convert = v.to("dyne cm^-2 s")
    assert convert.value == pytest.approx(10.0, DELTA)
    assert convert.unit == "dyne cm^-2 s"


def test_to_16():
    v = q.Quantity(1.0, "Pa s")
    convert = v.to("slug in^-1 s^-1")
    assert convert.value == pytest.approx(0.00174045320, DELTA)
    assert convert.unit == "slug in^-1 s^-1"


def test_to_17():
    v = q.Quantity(1.0, "kg m^-1 s^-1")
    convert = v.to("slug in^-1 s^-1")
    assert convert.value == pytest.approx(0.00174045320, DELTA)
    assert convert.unit == "slug in^-1 s^-1"


def test_to_18():
    v = q.Quantity(1.0, "lb ft^-1 s^-1")
    convert = v.to("Pa s")
    assert convert.value == pytest.approx(1.488164, DELTA)
    assert convert.unit == "Pa s"


def test_to_19():
    v = q.Quantity(1.0, "lb ft^-1 s^-1")
    convert = v.to("kg m^-1 s^-1")
    assert convert.value == pytest.approx(1.488164, DELTA)
    assert convert.unit == "kg m^-1 s^-1"


def test_to_20():
    v = q.Quantity(1.0, "Hz")
    with pytest.raises(ValueError) as e:
        convert = v.to("radian s^-1")
    assert e.value.from_unit == "Hz"
    assert e.value.to_unit == "radian s^-1"


def test_to_21():
    v = q.Quantity(1.0, "radian s^-1")
    with pytest.raises(ValueError) as e:
        convert = v.to("Hz")
    assert e.value.from_unit == "radian s^-1"
    assert e.value.to_unit == "Hz"


def test_to_22():
    v = q.Quantity(1.0, "lbf ft^-2")
    convert = v.to("N m^-2")
    assert convert.value == pytest.approx(47.88024159, DELTA)
    assert convert.unit == "N m^-2"


def test_to_23():
    v = q.Quantity(1.0, "ft^-3 s^-1")
    convert = v.to("m^-3 s^-1")
    assert convert.value == pytest.approx(35.3146667, DELTA)
    assert convert.unit == "m^-3 s^-1"


def test_to_24():
    v = q.Quantity(1.0, "m^-2")
    convert = v.to("cm^-2")
    assert convert.value == pytest.approx(0.0001, DELTA)
    assert convert.unit == "cm^-2"


def test_to_25():
    v = q.Quantity(1.0, "m^2")
    convert = v.to("in^2")
    assert convert.value == pytest.approx(1550.0031, DELTA)
    assert convert.unit == "in^2"


def test_to_26():
    v = q.Quantity(1.0, "radian s^-1")
    convert = v.to("degree s^-1")
    assert convert.value == pytest.approx(57.295779, DELTA)
    assert convert.unit == "degree s^-1"


def test_to_27():
    v = q.Quantity(1.0, "degree s^-1")
    convert = v.to("radian s^-1")
    assert convert.value == pytest.approx(0.01745329251, DELTA)
    assert convert.unit == "radian s^-1"


def test_to_28():
    v = q.Quantity(1.0, "dyne cm^-2")
    convert = v.to("N m^-2")
    assert convert.value == pytest.approx(0.1, DELTA)
    assert convert.unit == "N m^-2"


def test_to_29():
    v = q.Quantity(1.0, "psi")
    convert = v.to("Pa")
    assert convert.value == pytest.approx(6894.76, DELTA)
    assert convert.unit == "Pa"


def test_to_30():
    v = q.Quantity(1.0, "pdl")
    convert = v.to("N")
    assert convert.value == pytest.approx(0.138254999, DELTA)
    assert convert.unit == "N"


def test_to_31():
    v = q.Quantity(1.0, "ohm cm")
    convert = v.to("ohm m")
    assert convert.value == pytest.approx(0.01, DELTA)
    assert convert.unit == "ohm m"


def test_to_32():
    v = q.Quantity(1.0, "erg")
    convert = v.to("J")
    assert convert.value == pytest.approx(1.0e-7, DELTA)
    assert convert.unit == "J"


def test_to_33():
    v = q.Quantity(1.0, "BTU")
    convert = v.to("J")
    assert convert.value == pytest.approx(1055.056, DELTA)
    assert convert.unit == "J"


def test_to_34():
    v = q.Quantity(1.0, "gal")
    convert = v.to("m^3")
    assert convert.value == pytest.approx(0.00378541, DELTA)
    assert convert.unit == "m^3"


def test_to_35():
    v = q.Quantity(1.0, "l")
    convert = v.to("m^3")
    assert convert.value == pytest.approx(0.001, DELTA)
    assert convert.unit == "m^3"


def test_to_36():
    v = q.Quantity(1.0, "BTU lb^-1 R^-1")
    convert = v.to("J kg^-1 K^-1")
    assert convert.value == pytest.approx(4186.8161854, DELTA)
    assert convert.unit == "J kg^-1 K^-1"


def test_to_37():
    v = q.Quantity(1.0, "BTU lb^-1 F^-1")
    convert = v.to("J kg^-1 K^-1")
    assert convert.value == pytest.approx(4186.8161854, DELTA)
    assert convert.unit == "J kg^-1 K^-1"


def test_to_38():
    v = q.Quantity(1.0, "gal^-1")
    convert = v.to("m^-3")
    assert convert.value == pytest.approx(264.172, DELTA)
    assert convert.unit == "m^-3"


def test_to_39():
    v = q.Quantity(1.0, "BTU ft^-2")
    convert = v.to("J m^-2")
    assert convert.value == pytest.approx(11356.5713242, DELTA)
    assert convert.unit == "J m^-2"


def test_prop_constants_40():
    c = q.Quantity(10.5, "s^-0.5")
    x = q.Quantity(5.7, "m")
    t = q.Quantity(4.8, "s")

    y = c * x * t
    print(y)
    assert float(y) == 287.28


def test_unit_system_41():
    unitSysSI = q.UnitSystem("SI")
    myVel = q.Quantity(3.0, "ft s^-1")
    siVel = unitSysSI.convert(myVel)  # Returns velocity in m/s
    assert siVel.value == pytest.approx(0.9143999, DELTA)
    assert siVel.unit == "m s^-1"

    unitSysCGS = q.UnitSystem("CGS")
    cgsVel = unitSysCGS.convert(myVel)  # Returns velocity in cm/s
    assert cgsVel.value == 91.44
    assert cgsVel.unit == "cm s^-1"

    unitSysBT = q.UnitSystem("BT")
    btuVel = unitSysBT.convert(myVel)  # Returns velocity in ft/s
    assert btuVel.value == 3.0
    assert btuVel.unit == "ft s^-1"


def test_equality_42():
    qt1 = q.Quantity(5.0, "m s^-1")
    qt2 = q.Quantity(5.0, "m s^-1")
    assert qt1 == qt2


def test_math_fun_43():
    deg = q.Quantity(90, "degree")
    assert math.sin(deg) == 1.0

    rad = q.Quantity(math.pi / 2, "radian")
    assert math.sin(rad) == 1.0

    root = q.Quantity(100.0, "")
    assert math.sqrt(root) == 10.0


def test_subtraction_44():
    q1 = q.Quantity(10.0, "m s^-1")
    q2 = q.Quantity(5.0, "m s^-1")

    assert float(q1 - q2) == 5.0
    assert float(q2 - q1) == -5.0
    assert float(q1) - 2.0 == 8.0
    assert 2.0 - float(q1) == -8.0
    assert float(q1) - 3 == 7.0
    assert 3 - float(q1) == -7.0


def test_pow_45():
    q1 = q.Quantity(10.0, "m s^-1")
    q2 = q.Quantity(5.0, "m s^-1")

    q1_sq = q1**2
    assert q1_sq.unit == "m^2 s^-2"

    assert float(q1) ** 2 == 100.0
    assert float(q2) ** 2 == 25.0


def test_eq_46():
    q1 = q.Quantity(10.0, "m s^-1")
    q2 = q.Quantity(5.0, "m s^-1")
    q3 = q.Quantity(10.0, "m s^-1")
    q4 = q.Quantity(10.0, "")

    assert q1 != q2
    assert q1 == q3
    assert float(q1) == 10.0
    assert q4 == 10.0


def test_rdiv_47():
    q1 = q.Quantity(10.0, "m s^-1")
    q2 = q.Quantity(5.0, "m s^-1")

    assert float(q1) / float(q2) == 2.0
    assert float(q2) / float(q1) == 0.5
    assert float(q1) / 2 == 5.0
    assert 2.0 / float(q1) == 0.2


def test_tempK_48():
    k = q.Quantity(-40, "K")

    kc = k.to("C")
    assert kc.value == -313.15
    assert kc.unit == "C"

    kc = k.to("R")
    assert kc.value == -72.0
    assert kc.unit == "R"

    kc = k.to("F")
    assert kc.value == pytest.approx(-531.67, DELTA)
    assert kc.unit == "F"


def test_temp_49():
    mk = q.Quantity(-40_000, "mK")
    uc = mk.to("uC^1")
    assert uc.value == -3.13150000e08


def test_temp_50():
    k = q.Quantity(1.0, "K")

    f = k.to("F")
    r = k.to("R")
    c = k.to("C")

    assert f.value == -457.87
    assert r.value == 1.8
    assert c.value == -272.15


def test_temp_51():
    c = q.Quantity(1.0, "C")

    f = c.to("F")
    r = c.to("R")
    k = c.to("K")

    assert f.value == 33.80
    assert r.value == pytest.approx(493.469, DELTA)
    assert k.value == 274.15


def test_temp_52():
    r = q.Quantity(1.0, "R")

    f = r.to("F")
    c = r.to("C")
    k = r.to("K")

    assert f.value == pytest.approx(-458.6699, DELTA)
    assert c.value == pytest.approx(-272.5944, DELTA)
    assert k.value == pytest.approx(0.555556, DELTA)


def test_temp_53():
    f = q.Quantity(1.0, "F")

    c = f.to("C")
    r = f.to("R")
    k = f.to("K")

    assert c.value == pytest.approx(-17.2222, DELTA)
    assert r.value == pytest.approx(460.670, DELTA)
    assert k.value == pytest.approx(255.927, DELTA)


def test_temp_54():
    hc = q.Quantity(1.0, "J g^-1 K^-1")

    hcto1 = hc.to("kJ kg^-1 K^-1")

    assert hcto1.value == pytest.approx(1.0, DELTA)
    assert hcto1.unit == "kJ kg^-1 K^-1"

    hcto2 = hc.to("J kg^-1 C^-1")

    assert hcto2.value == pytest.approx(1000.0, DELTA)
    assert hcto2.unit == "J kg^-1 C^-1"

    hcto3 = hc.to("kJ kg^-1 C^-1")

    assert hcto3.value == pytest.approx(1.0, DELTA)
    assert hcto3.unit == "kJ kg^-1 C^-1"

    hcto4 = hc.to("cal g^-1 C^-1")

    assert hcto4.value == pytest.approx(0.2390057, DELTA)
    assert hcto4.unit == "cal g^-1 C^-1"

    hcto5 = hc.to("cal kg^-1 C^-1")

    assert hcto5.value == pytest.approx(239.0057, DELTA)
    assert hcto5.unit == "cal kg^-1 C^-1"

    hcto6 = hc.to("kcal kg^-1 C^-1")

    assert hcto6.value == pytest.approx(0.2390057, DELTA)
    assert hcto6.unit == "kcal kg^-1 C^-1"

    hcto7 = hc.to("BTU lb^-1 F^-1")

    assert hcto7.value == pytest.approx(0.238845, DELTA)
    assert hcto7.unit == "BTU lb^-1 F^-1"


def test_temp_54():
    temp_var = q.Quantity(1.0, "kg m^-3 s^-1 K^2")

    temp_varto1 = temp_var.to("g cm^-3 s^-1 K^2")

    assert temp_varto1.value == pytest.approx(0.001, DELTA)
    assert temp_varto1.unit == "g cm^-3 s^-1 K^2"

    temp_varto2 = temp_var.to("kg mm^-3 s^-1 K^2")

    assert temp_varto2.value == pytest.approx(1e-09, DELTA)
    assert temp_varto2.unit == "kg mm^-3 s^-1 K^2"

    temp_varto3 = temp_var.to("kg um^-3 s^-1 K^2")

    assert temp_varto3.value == pytest.approx(9.999999999999999e-19, DELTA)
    assert temp_varto3.unit == "kg um^-3 s^-1 K^2"

    temp_varto4 = temp_var.to("mg mm^-3 ms^-1 K^2")

    assert temp_varto4.value == pytest.approx(1.0000000000000002e-06, DELTA)
    assert temp_varto4.unit == "mg mm^-3 ms^-1 K^2"

    temp_varto5 = temp_var.to("g cm^-3 us^-1 K^2")

    assert temp_varto5.value == pytest.approx(1e-09, DELTA)
    assert temp_varto5.unit == "g cm^-3 us^-1 K^2"

    temp_varto6 = temp_var.to("pg um^-3 ms^-1 K^2")

    assert temp_varto6.value == pytest.approx(9.999999999999997e-07, DELTA)
    assert temp_varto6.unit == "pg um^-3 ms^-1 K^2"


def test_power_56():
    qt = q.Quantity(5.0, "m^0")
    qtm = qt * 2

    assert qtm.value == 10.0
    assert qtm.unit == ""


def test_ge_57():
    x = q.Quantity(10.5, "cm")
    y = q.Quantity(10.5, "m")
    z = q.Quantity(10.5, "g")
    r = q.Quantity(10.5, "")

    assert y >= x
    assert 15.7 >= r
    assert r >= 7.8

    with pytest.raises(ValueError) as e_info:
        assert x >= z
        assert x >= y
        assert 5.0 >= r

    with pytest.raises(TypeError) as e_info:
        assert x >= 5.0


def test_gt_59():
    x = q.Quantity(10.5, "cm")
    y = q.Quantity(10.5, "m")
    z = q.Quantity(10.5, "g")
    r = q.Quantity(10.5, "")

    assert y > x
    assert 15.7 > r
    assert r > 7.8

    with pytest.raises(ValueError) as e_info:
        assert x > z
        assert x > y
        assert 5.0 > r

    with pytest.raises(TypeError) as e_info:
        assert x > 5.0


def test_lt_60():
    x = q.Quantity(10.5, "cm")
    y = q.Quantity(10.5, "m")
    z = q.Quantity(10.5, "g")
    r = q.Quantity(10.5, "")

    assert x < y
    assert r < 15.7
    assert 7.8 < r

    with pytest.raises(ValueError) as e_info:
        assert z < x
        assert y < x
        assert r < 0.5

    with pytest.raises(TypeError) as e_info:
        assert 5.0 < x


def test_le_61():
    x = q.Quantity(10.5, "cm")
    y = q.Quantity(10.5, "m")
    z = q.Quantity(10.5, "g")
    r = q.Quantity(10.5, "")

    assert x <= y
    assert r <= 15.7
    assert 7.8 <= r

    with pytest.raises(ValueError) as e_info:
        assert z <= x
        assert y <= x
        assert r <= 0.5

    with pytest.raises(TypeError) as e_info:
        assert 5.0 <= x


def test_eq_62():
    x = q.Quantity(10.5, "cm")
    y = q.Quantity(10.5, "m")
    z = q.Quantity(10.5, "g")
    r = q.Quantity(10.5, "")

    l = q.Quantity(10.5, "cm")
    m = q.Quantity(10.5, "m")
    n = q.Quantity(10.5, "")

    assert x == l
    assert y == m
    assert r == n

    with pytest.raises(ValueError) as e_info:
        assert z == x
        assert y == x
        assert r == 0.5

    with pytest.raises(TypeError) as e_info:
        assert 5.0 == x


def test_neq_63():
    x = q.Quantity(10.5, "cm")
    y = q.Quantity(10.5, "m")
    z = q.Quantity(10.5, "g")
    r = q.Quantity(10.5, "")

    assert y != x
    assert x != y

    assert r != 0.5
    assert 0.5 != r


def test_temp_inverse_64():
    c = q.Quantity(2.0, "C")
    assert float(c) == 275.15

    c_inverse = q.Quantity(2.0, "C^-1")
    assert float(c_inverse) == 2.0


def test_temp_inverse_65():
    f = q.Quantity(2.0, "F")
    assert float(f) == pytest.approx(256.483311, DELTA)

    f_inverse = q.Quantity(2.0, "F^-1")
    assert float(f_inverse) == pytest.approx(3.5999999999999996, DELTA)


def test_temp_type_66():
    c0 = q.Quantity(1.0, "C")
    assert c0.type == "Temperature"

    c1 = q.Quantity(1.0, "J kg^-1 C^-1")
    assert c1.type == "Temperature Difference"

    c2 = q.Quantity(1.0, "kg m^-3 s^-1 K^2")
    assert c2.type == "Temperature Difference"

    c4 = q.Quantity(1.0, "F")
    assert c4.type == "Temperature"

    c6 = q.Quantity(1.0, "F^1")
    assert c6.type == "Temperature Difference"

    c7 = q.Quantity(1.0, "F^-1")
    assert c7.type == "Temperature Difference"

    c8 = q.Quantity(1.0, "F^2")
    assert c8.type == "Temperature Difference"


def test_temp_difference_67():
    td1 = q.Quantity(150.0, "delta_C")
    assert td1.type == "Temperature Difference"

    td2 = q.Quantity(100.0, "delta_C")
    assert td2.type == "Temperature Difference"

    td = td1 - td2
    assert td.type == "Temperature Difference"

    td_m = td * 2
    assert td_m.unit == "delta_K"
    assert td_m.type == "Temperature Difference"

    t1 = q.Quantity(150.0, "C")
    assert t1.type == "Temperature"

    t2 = q.Quantity(100.0, "C")
    assert t2.type == "Temperature"

    td = t1 - t2
    assert td.type == "Temperature Difference"

    td2 = t2 - t1
    assert td2.type == "Temperature Difference"

    tc1 = q.Quantity(100.0, "C")
    td1 = q.Quantity(50.0, "C^-1")

    with pytest.raises(ValueError) as e:
        t = tc1 + td1


def test_core_temp_68():
    t1 = q.Quantity(1.0, "K")
    assert float(t1) == 1.0
    assert t1.type == "Temperature"

    t2 = q.Quantity(2.0, "K")
    assert float(t2) == 2.0
    assert t2.type == "Temperature"

    dt1 = t2 - t1
    assert float(dt1) == 1.0
    assert dt1.type == "Temperature Difference"

    t3 = q.Quantity(1.0, "C")
    assert float(t3) == 274.15
    assert t3.type == "Temperature"

    t4 = q.Quantity(2.0, "C")
    assert float(t4) == 275.15
    assert t4.type == "Temperature"

    dt2 = t4 - t3
    assert float(dt2) == 1.0
    assert dt2.type == "Temperature Difference"

    invt1 = q.Quantity(1.0, "K^-1")
    assert float(invt1) == 1.0
    assert invt1.type == "Temperature Difference"

    dt3 = 1.0 / invt1
    assert float(dt3) == 1.0
    assert dt1.type == dt2.type == dt3.type

    invt2 = q.Quantity(1.0, "C^-1")
    assert float(invt2) == 1.0
    assert invt2.type == "Temperature Difference"

    dt4 = 1.0 / invt2
    assert float(dt4) == 1.0
    assert dt4.type == "Temperature Difference"


def test_temp_addition_69():
    t1 = q.Quantity(150.0, "C")
    t2 = q.Quantity(50.0, "C")

    td = t1 - t2
    assert td.type == "Temperature Difference"
    assert float(td) == 100.0
    assert td.unit == "delta_K"

    kd = q.Quantity(50.0, "delta_C")
    k = q.Quantity(50.0, "K")

    t = k + kd
    assert float(t) == 100.0
    assert t.type == "Temperature"


def test_quantity_map_70():
    quantity_map_from_settings_API = {
        "Mass": 1,
        "Velocity": 2.5,
        "Current": 3,
        "Light": 1,
        "Epsilon Flux Coefficient": 2,
    }

    api_test = q.Quantity(10.5, quantity_map=quantity_map_from_settings_API)
    assert api_test.value == 10.5
    assert api_test.unit == "kg^3 m^-1.5 s^-6.5 A^3 cd"


def test_quantity_map_71():
    quantity_map_from_settings_API = {
        "Mass": 1,
        "Velocity": 2.5,
        "Current": 3,
        "Light": 2,
        "Epsilon Flux Coefficient": 2,
    }

    with pytest.raises(ValueError):
        api_test = q.Quantity(
            10.5, unit_str="kg m s^-1", quantity_map=quantity_map_from_settings_API
        )


def test_quantity_map_72():
    quantity_map_from_settings_API = {
        "Temperature": 1,
        "Pressure": 1,
        "Volume": 1,
    }

    api_test = q.Quantity(10.5, quantity_map=quantity_map_from_settings_API)
    assert api_test.value == 10.5
    assert api_test.unit == "K Pa m^3"


def test_unit_from_dimensions_73():
    p = q.Quantity(10.5, dimensions=[1.0, -1.0, -2.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
    assert p.unit == "kg m^-1 s^-2"


def test_unit_from_dimensions_74():
    l = q.Quantity(10.5, dimensions=[0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
    assert l.unit == "m"


def test_unit_from_dimensions_75():
    x = q.Quantity(10.5, dimensions=[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
    assert x.unit == ""


def test_unit_from_dimensions_76():
    test = q.Quantity(10.5, dimensions=[0, 1, -1])
    assert test.unit == "m s^-1"
    assert test.get_dimensions_list() == [0.0, 1.0, -1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]


def test_unit_from_dimensions_77():
    test = q.Quantity(10.5, dimensions=[0, 1.0, -2.0])
    assert test.unit == "m s^-2"
    assert test.get_dimensions_list() == [0.0, 1.0, -2.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]


def testing_dimensions():
    print(f"{'*' * 25} {testing_dimensions.__name__} {'*' * 25}")

    def dim_test(unit_str, dim_list):
        qt = q.Quantity(10, unit_str)
        print(f"{unit_str} : {qt.get_dimensions_list()}")
        assert qt.get_dimensions_list() == dim_list

    dim_test("m", [0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
    dim_test("m s^-1", [0.0, 1.0, -1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
    dim_test("kg m s^-2 m^-2", [1.0, -1.0, -2.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
    dim_test("Pa", [1.0, -1.0, -2.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
    dim_test("kPa", [1.0, -1.0, -2.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
    dim_test("Pa^2", [2.0, -2.0, -4.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
    dim_test("daPa", [1.0, -1.0, -2.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
    dim_test("MPa", [1.0, -1.0, -2.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
    dim_test("kPa^2", [2.0, -2.0, -4.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
    dim_test("slug in^-1 s^-1", [1.0, -1.0, -1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
    dim_test("radian", [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0])
    dim_test("ohm", [1.0, 2.0, -3.0, 0.0, 0.0, 0.0, 0.0, -2.0, 0.0])
    dim_test("lb cm s^-2", [1.0, 1.0, -2.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
    print("-" * 75)


def testing_to_systems():
    print(f"{'*' * 25} {testing_to_systems.__name__} {'*' * 25}")
    test = q.Quantity(90, "lb cm s^-2")
    print(f"test : {test.get_dimensions_list()}")  # [1 1 -2 0 -2 0 0 0 0]

    test.to("kg m s^-2")
    test.to("g m s^-2")
    test.to("g cm s^-2")
    test.to("g in s^-2")
    test.to("g ft s^-2")
    test.to("kg ft s^-2")
    test.to("kg in s^-1 s^-1")

    with pytest.raises(ValueError) as e:
        test.to("ft s^-2")
    print("-" * 75)


def testing_multipliers():
    print(f"{'*' * 25} {testing_multipliers.__name__} {'*' * 25}")

    def from_to(from_str, to_str):
        qt = q.Quantity(1, from_str)
        to = qt.to(to_str)
        print(f"from {qt} -> to {to}")

    from_to("mm", "cm")
    from_to("m", "ft")
    from_to("dm^3", "m^3")
    from_to("m s^-1", "cm s^-1")
    from_to("N", "dyne")
    from_to("m^2", "in^2")
    from_to("degree s^-1", "radian s^-1")
    from_to("radian s^-1", "degree s^-1")
    from_to("Pa", "lb m s^-2 ft^-2")
    from_to("lb m s^-2 ft^-2", "Pa")

    from_to("J kg^-1 K^-1", "J kg^-1 C^-1")
    from_to("J kg^-1 K^-1", "J kg^-1 R^-1")
    from_to("J kg^-1 K^-1", "J kg^-1 F^-1")

    from_to("K", "C")
    from_to("K", "R")
    from_to("K", "F")

    print("-" * 75)


def testing_arithmetic_operators():
    print(f"{'*' * 25} {testing_arithmetic_operators.__name__} {'*' * 25}")

    qt1 = q.Quantity(10, "m s^-1")
    qt2 = q.Quantity(5, "m s^-1")

    qt3 = qt1 * qt2

    print(f"{qt1} * {qt2} =  {qt3}")
    assert qt3.value == 50
    assert qt3.unit == "m^2 s^-2"

    result = qt1 * 2
    print(f"{qt1} * {2} =  {result}")
    assert result.value == 20
    assert result.unit == "m s^-1"

    result1 = 2 * qt1
    print(f"{2} * {qt1} =  {result1}")
    assert result1.value == 20
    assert result1.unit == "m s^-1"

    q3 = qt1 / qt2

    print(f"{qt1} / {qt2} =  {q3}")
    assert q3.value == 2
    assert q3.unit == ""

    result3 = qt1 / 2
    print(f"{qt1} / {2} =  {qt1 / 2}")
    assert result3.value == 5
    assert result3.unit == "m s^-1"

    qa3 = qt1 + qt2

    print(f"{qt1} + {qt2} =  {qa3}")
    assert qa3.value == 15
    assert qa3.unit == "m s^-1"

    with pytest.raises(TypeError) as e:
        result5 = qt1 + 2
        print(f"{qt1} + {2} =  {result5}")

    with pytest.raises(ValueError) as e:
        result6 = 2 + qt1
        print(f"{2} + {qt1} =  {result6}")

    qs3 = qt1 - qt2

    print(f"{qt1} - {qt2} =  {qs3}")
    assert qs3.value == 5
    assert qs3.unit == "m s^-1"

    with pytest.raises(TypeError) as e:
        result7 = qt1 - 2
        print(f"{qt1} - {2} =  {result7}")

    with pytest.raises(ValueError) as e:
        result8 = 2 - qt1
        print(f"{2} - {qt1} =  {result8}")


def testing_properties():
    print(f"{'*' * 25} {testing_properties.__name__} {'*' * 25}")

    v = q.Quantity(1, "cm s^-1")
    print(f"value = {v.value}")
    print(f"unit = {v.unit}")
    print(f"si value = {v._si_value}")
    print(f"si unit = {v._si_unit}")
    print(f"is dimensionless? = {v.is_dimensionless()}")
    print(f"dimensions = {v.get_dimensions_list()}")


# if __name__ == "__main__":
# test_value_unit_1()
# testing_dimensions()
# testing_multipliers()
# testing_to_systems()
# testing_arithmetic_operators()
# testing_properties()
#
# x = q.Quantity(1, "ft")
# print(
#     f"User unit: {x._unit.user_unit}, multiplier: {x._unit.si_factor}, reduced_si_unit: {x._unit.si_unit}, si_value: {x._si_value}"
# )



#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import _ChildNamedObjectAccessorMixin

from ansys.fluent.core.solver.flobject import _CreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _NonCreatableNamedObjectMixin

from .subspace_size import subspace_size
from .skip_itr import skip_itr
class reduced_rank_extrapolation_options(Group):
    """
    'reduced_rank_extrapolation_options' child.
    """

    fluent_name = "reduced-rank-extrapolation-options"

    child_names = \
        ['subspace_size', 'skip_itr']

    subspace_size: subspace_size = subspace_size
    """
    subspace_size child of reduced_rank_extrapolation_options.
    """
    skip_itr: skip_itr = skip_itr
    """
    skip_itr child of reduced_rank_extrapolation_options.
    """



"""Provides a module for launching Fluent.

This module supports both starting Fluent locally and connecting to a
remote instance with gRPC.
"""
from enum import Enum
import json
import logging
import os
from pathlib import Path
import platform
import subprocess
import tempfile
import time
from typing import Any, Dict, Union
import warnings

from ansys.fluent.core.fluent_connection import FluentConnection
from ansys.fluent.core.launcher.fluent_container import start_fluent_container
from ansys.fluent.core.scheduler import build_parallel_options, load_machines
from ansys.fluent.core.session import _parse_server_info_file
from ansys.fluent.core.session_meshing import Meshing
from ansys.fluent.core.session_pure_meshing import PureMeshing
from ansys.fluent.core.session_solver import Solver
from ansys.fluent.core.session_solver_icing import SolverIcing
from ansys.fluent.core.utils.networking import find_remoting_ip
import ansys.platform.instancemanagement as pypim

_THIS_DIR = os.path.dirname(__file__)
_OPTIONS_FILE = os.path.join(_THIS_DIR, "fluent_launcher_options.json")
logger = logging.getLogger("ansys.fluent.launcher")


def _is_windows():
    """Check if the current operating system is windows."""
    return platform.system() == "Windows"


class FluentVersion(Enum):
    """An enumeration over supported Fluent versions."""

    version_24R1 = "24.1.0"
    version_23R2 = "23.2.0"
    version_23R1 = "23.1.0"
    version_22R2 = "22.2.0"

    @classmethod
    def _missing_(cls, version):
        if isinstance(version, (float, str)):
            version = str(version) + ".0"
            for v in FluentVersion:
                if version == v.value:
                    return FluentVersion(version)
            else:
                raise RuntimeError(
                    f"The specified version '{version[:-2]}' is not supported."
                    + f" Supported versions are: "
                    + ", ".join([ver.value for ver in FluentVersion][::-1])
                )

    def __str__(self):
        return str(self.value)


def get_ansys_version() -> str:
    """Return the version string corresponding to the most recent, available ANSYS
    installation. The returned value is the string component of one of the members
    of the FluentVersion class.
    """
    for v in FluentVersion:
        if "AWP_ROOT" + "".join(str(v).split("."))[:-1] in os.environ:
            return str(v)

    raise RuntimeError("No ANSYS version can be found.")


def get_fluent_exe_path(**launch_argvals) -> Path:
    """Get Fluent executable path. The path is searched in the following order.

    1. ``product_version`` parameter passed with ``launch_fluent``.
    2. The latest ANSYS version from ``AWP_ROOTnnn``` environment variables.

    Returns
    -------
    Path
        Fluent executable path
    """

    def get_fluent_root(version: FluentVersion) -> Path:
        awp_root = os.environ["AWP_ROOT" + "".join(str(version).split("."))[:-1]]
        return Path(awp_root) / "fluent"

    def get_exe_path(fluent_root: Path) -> Path:
        if _is_windows():
            return fluent_root / "ntbin" / "win64" / "fluent.exe"
        else:
            return fluent_root / "bin" / "fluent"

    # (DEV) "PYFLUENT_FLUENT_ROOT" environment variable
    fluent_root = os.getenv("PYFLUENT_FLUENT_ROOT")
    if fluent_root:
        return get_exe_path(Path(fluent_root))

    # Look for Fluent exe path in the following order:
    # 1. product_version parameter passed with launch_fluent
    product_version = launch_argvals.get("product_version")
    if product_version:
        return get_exe_path(get_fluent_root(FluentVersion(product_version)))

    # 2. the latest ANSYS version from AWP_ROOT environment variables
    ansys_version = get_ansys_version()
    return get_exe_path(get_fluent_root(FluentVersion(ansys_version)))


class LaunchMode(Enum):
    """An enumeration over supported launch modes."""

    # Tuple: Name, Solver object type, Meshing flag, Launcher options
    MESHING_MODE = ("meshing", Meshing, True, [])
    PURE_MESHING_MODE = ("pure-meshing", PureMeshing, True, [])
    SOLVER = ("solver", Solver, False, [])
    SOLVER_ICING = ("solver-icing", SolverIcing, False, [("fluent_icing", True)])

    @staticmethod
    def get_mode(mode: str):
        """Returns the LaunchMode based on the provided mode string."""
        for m in LaunchMode:
            if mode == m.value[0]:
                return m
        else:
            raise RuntimeError(
                f"The passed mode '{mode}' matches none of the allowed modes."
            )


def _get_server_info_filepath():
    server_info_dir = os.getenv("SERVER_INFO_DIR")
    dir_ = Path(server_info_dir) if server_info_dir else tempfile.gettempdir()
    fd, filepath = tempfile.mkstemp(suffix=".txt", prefix="serverinfo-", dir=str(dir_))
    os.close(fd)
    return filepath


def _get_subprocess_kwargs_for_fluent(env: Dict[str, Any]) -> Dict[str, Any]:
    kwargs: Dict[str, Any] = {}
    kwargs.update(stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    if _is_windows():
        kwargs.update(creationflags=subprocess.CREATE_NEW_PROCESS_GROUP)
    else:
        kwargs.update(shell=True, start_new_session=True)
    fluent_env = os.environ.copy()
    fluent_env.update({k: str(v) for k, v in env.items()})
    fluent_env["REMOTING_THROW_LAST_TUI_ERROR"] = "1"
    from ansys.fluent.core import INFER_REMOTING_IP

    if INFER_REMOTING_IP and not "REMOTING_SERVER_ADDRESS" in fluent_env:
        remoting_ip = find_remoting_ip()
        if remoting_ip:
            fluent_env["REMOTING_SERVER_ADDRESS"] = remoting_ip
    kwargs.update(env=fluent_env)
    return kwargs


def _build_fluent_launch_args_string(**kwargs) -> str:
    """Build Fluent's launch arguments string from keyword arguments.

    Returns
    -------
    str
        Fluent's launch arguments string.
    """
    all_options = None
    with open(_OPTIONS_FILE, encoding="utf-8") as fp:
        all_options = json.load(fp)
    launch_args_string = ""
    for k, v in all_options.items():
        argval = kwargs.get(k)
        default = v.get("default")
        if argval is None and v.get("fluent_required") is True:
            argval = default
        if argval is not None:
            allowed_values = v.get("allowed_values")
            if allowed_values and argval not in allowed_values:
                if default is not None:
                    old_argval = argval
                    argval = default
                    logger.warning(
                        "Default value %s is chosen for %s as the passed "
                        "value  %s is outside allowed values %s.",
                        argval,
                        k,
                        old_argval,
                        allowed_values,
                    )
                else:
                    logger.warning(
                        "%s = %s is discarded as it is outside " "allowed values %s.",
                        k,
                        argval,
                        allowed_values,
                    )
                    continue
            fluent_map = v.get("fluent_map")
            if fluent_map:
                if isinstance(argval, str):
                    json_key = argval
                else:
                    json_key = json.dumps(argval)
                argval = fluent_map[json_key]
            launch_args_string += v["fluent_format"].replace("{}", str(argval))
    addArgs = kwargs["additional_arguments"]
    if "-t" not in addArgs and "-cnf=" not in addArgs:
        mlist = load_machines(ncores=kwargs["processor_count"])
        launch_args_string += " " + build_parallel_options(mlist)
    return launch_args_string


def launch_remote_fluent(
    session_cls,
    start_transcript: bool,
    start_timeout: int = 100,
    product_version: str = None,
    cleanup_on_exit: bool = True,
    meshing_mode: bool = False,
    dimensionality: str = None,
    launcher_args: Dict[str, Any] = None,
) -> Union[Meshing, PureMeshing, Solver, SolverIcing]:
    """Launch Fluent remotely using `PyPIM <https://pypim.docs.pyansys.com>`.

    When calling this method, you must ensure that you are in an
    environment where PyPIM is configured. You can use the :func:
    `pypim.is_configured <ansys.platform.instancemanagement.is_configured>`
    method to verify that PYPIM is configured.

    Parameters
    ----------
    session_cls: Union[type(Meshing), type(PureMeshing), type(Solver), type(SolverIcing)]
        Session type.
    start_transcript: bool
        Whether to start streaming the Fluent transcript in the client. The
        default is ``True``. You can stop and start the streaming of the
        Fluent transcript subsequently via method calls on the session object.
    start_timeout : int, optional
        Maximum allowable time in seconds for connecting to the Fluent
        server. The default is ``100``.
    product_version : str, optional
        Select an installed version of ANSYS. The string must be in a format like
        ``"23.1.0"`` (for 2023 R1) matching the documented version format in the
        FluentVersion class. The default is ``None``, in which case the newest installed
        version is used.
    cleanup_on_exit : bool, optional
        Whether to clean up and exit Fluent when Python exits or when garbage
        is collected for the Fluent Python instance. The default is ``True``.
    meshing_mode : bool, optional
        Whether to launch Fluent remotely in meshing mode. The default is
        ``False``.
    dimensionality : str, optional
        Geometric dimensionality of the Fluent simulation. The default is ``None``,
        in which case ``"3d"`` is used. Options are ``"3d"`` and ``"2d"``.

    Returns
    -------
    Union[Meshing, PureMeshing, Solver, SolverIcing]
        Session object.
    """
    pim = pypim.connect()
    instance = pim.create_instance(
        product_name="fluent-meshing"
        if meshing_mode
        else "fluent-2ddp"
        if dimensionality == "2d"
        else "fluent-3ddp",
        product_version=product_version,
    )
    instance.wait_for_ready()
    # nb pymapdl sets max msg len here:
    channel = instance.build_grpc_channel()
    return session_cls(
        fluent_connection=FluentConnection(
            channel=channel,
            cleanup_on_exit=cleanup_on_exit,
            remote_instance=instance,
            start_timeout=start_timeout,
            start_transcript=start_transcript,
            launcher_args=launcher_args,
        )
    )


def _get_session_info(argvals, mode: Union[LaunchMode, str, None] = None):
    """Updates the session information."""
    if mode is None:
        mode = LaunchMode.SOLVER

    if isinstance(mode, str):
        mode = LaunchMode.get_mode(mode)
    new_session = mode.value[1]
    meshing_mode = mode.value[2]
    for k, v in mode.value[3]:
        argvals[k] = v

    return new_session, meshing_mode, argvals, mode


def _raise_exception_g_gu_in_windows_os(additional_arguments: str) -> None:
    """If -g or -gu is passed in Windows OS, the exception should be raised."""
    additional_arg_list = additional_arguments.split()
    if _is_windows() and (
        ("-g" in additional_arg_list) or ("-gu" in additional_arg_list)
    ):
        raise ValueError("'-g' and '-gu' is not supported on windows platform.")


def _update_launch_string_wrt_gui_options(
    launch_string: str, show_gui: bool = None, additional_arguments: str = ""
) -> str:
    """Checks for all gui options in additional arguments and updates the
    launch string with hidden, if none of the options are met."""

    if (show_gui is False) or (
        show_gui is None and (os.getenv("PYFLUENT_SHOW_SERVER_GUI") != "1")
    ):
        if not {"-g", "-gu"} & set(additional_arguments.split()):
            launch_string += " -hidden"

    return launch_string


def _await_fluent_launch(
    server_info_filepath: str, start_timeout: int, sifile_last_mtime: float
):
    """Wait for successful fluent launch or raise an error."""
    while True:
        if Path(server_info_filepath).stat().st_mtime > sifile_last_mtime:
            time.sleep(1)
            logger.info("Fluent process is successfully launched.")
            break
        if start_timeout == 0:
            raise RuntimeError("The launch process has been timed out.")
        time.sleep(1)
        start_timeout -= 1
        logger.info(
            "Waiting for Fluent to launch...%02d seconds remaining",
            start_timeout,
        )


def _connect_to_running_server(argvals, server_info_filepath: str):
    """Connect to an already running session."""
    ip = argvals.get("ip", None)
    port = argvals.get("port", None)
    password = argvals.get("password", None)
    if ip and port:
        warnings.warn(
            "The server-info file was not parsed because ip and port were provided explicitly."
        )
    elif server_info_filepath:
        ip, port, password = _parse_server_info_file(server_info_filepath)
    elif os.getenv("PYFLUENT_FLUENT_IP") and os.getenv("PYFLUENT_FLUENT_PORT"):
        ip = port = None
    else:
        raise RuntimeError(
            "Please provide either ip and port data or server-info file."
        )

    return ip, port, password


def _get_running_session_mode(
    fluent_connection: FluentConnection, mode: LaunchMode = None
):
    """Get the mode of the running session if the mode has not been mentioned
    explicitly."""
    if mode:
        session_mode = mode
    else:
        try:
            session_mode = LaunchMode.get_mode(
                fluent_connection.get_current_fluent_mode()
            )
        except BaseException:
            raise RuntimeError("Fluent session password mismatch")
    return session_mode.value[1]


def _start_instance(start_instance: Union[bool, None]):
    """Sets up how to start an instance of fluent."""
    if start_instance is None:
        return bool(
            int(
                os.getenv(
                    "PYFLUENT_START_INSTANCE", "0" if pypim.is_configured() else "1"
                )
            )
        )
    return start_instance


def _generate_launch_string(
    argvals,
    meshing_mode: bool,
    show_gui: bool,
    additional_arguments: str,
    server_info_filepath: str,
):
    """Generates the launch string to launch fluent."""
    exe_path = get_fluent_exe_path(**argvals)
    launch_string = str(exe_path)
    launch_string += _build_fluent_launch_args_string(**argvals)
    if meshing_mode:
        launch_string += " -meshing"
    launch_string += f" {additional_arguments}"
    launch_string += f' -sifile="{server_info_filepath}"'
    launch_string += " -nm"
    launch_string = _update_launch_string_wrt_gui_options(
        launch_string, show_gui, additional_arguments
    )
    return launch_string


def scm_to_py(topy):
    """Convert journal filenames to Python filename."""
    if not isinstance(topy, (str, list)):
        raise TypeError("Journal name should be of str or list type.")
    if isinstance(topy, str):
        topy = [topy]
    fluent_jou_arg = "".join([f'-i "{journal}" ' for journal in topy])
    return f" {fluent_jou_arg} -topy"


class LaunchFluentError(Exception):
    """Exception class representing launch errors."""

    def __init__(self, launch_string):
        """__init__ method of LaunchFluentError class."""
        details = "\n" + "Fluent Launch string: " + launch_string
        super().__init__(details)


#   pylint: disable=unused-argument
def launch_fluent(
    product_version: str = None,
    version: str = None,
    precision: str = None,
    processor_count: int = None,
    journal_filepath: str = None,
    start_timeout: int = 100,
    additional_arguments: str = "",
    env: Dict[str, Any] = None,
    start_instance: bool = None,
    ip: str = None,
    port: int = None,
    cleanup_on_exit: bool = True,
    start_transcript: bool = True,
    show_gui: bool = None,
    case_filepath: str = None,
    case_data_filepath: str = None,
    lightweight_mode: bool = False,
    mode: Union[LaunchMode, str, None] = None,
    server_info_filepath: str = None,
    password: str = None,
    py: bool = None,
    gpu: bool = None,
    cwd: str = None,
    topy: Union[str, list] = None,
    **kwargs,
) -> Union[Meshing, PureMeshing, Solver, SolverIcing]:
    """Launch Fluent locally in server mode or connect to a running Fluent
    server instance.

    Parameters
    ----------
    product_version : str, optional
        Select an installed version of ANSYS. The string must be in a format like
        ``"23.1.0"`` (for 2023 R1) matching the documented version format in the
        FluentVersion class. The default is ``None``, in which case the newest installed
        version is used.
    version : str, optional
        Geometric dimensionality of the Fluent simulation. The default is ``None``,
        in which case ``"3d"`` is used. Options are ``"3d"`` and ``"2d"``.
    precision : str, optional
        Floating point precision. The default is ``None``, in which case ``"double"``
        is used. Options are ``"double"`` and ``"single"``.
    processor_count : int, optional
        Number of processors. The default is ``None``, in which case ``1``
        processor is used.  In job scheduler environments the total number of
        allocated cores is clamped to this value.
    journal_filepath : str, optional
        Name of the journal file to read. The default is ``None``.
    start_timeout : int, optional
        Maximum allowable time in seconds for connecting to the Fluent
        server. The default is ``100``.
    additional_arguments : str, optional
        Additional arguments to send to Fluent as a string in the same
        format they are normally passed to Fluent on the command line.
        The default is``""``.
    env : dict[str, str], optional
        Mapping to modify environment variables in Fluent. The default
        is ``None``.
    start_instance : bool, optional
        Whether to start a local Fluent instance. The default is None, which
        indicates True. Otherwise, connect to an existing Fluent instance at a
        specified IP address on a specified port, using the arguments ``ip`` and
        ``port``. You can also use the environment variable ``PYFLUENT_START_INSTANCE=<0 or 1>``
        to set ``start_instance`` if you do not pass it as an argument.
    ip : str, optional
        IP address for connecting to an existing Fluent instance. This parameter
        is used only when ``start_instance`` is ``False``. Otherwise, the
        IP address defaults to ``"127.0.0.1"``. You can also use the environment
        variable ``PYFLUENT_FLUENT_IP=<ip>`` to set this parameter.
    port : int, optional
        Port to listen on for an existing Fluent instance. This parameter is
        used only when ``start_instance`` is ``False``. You can use the
        environment variable ``PYFLUENT_FLUENT_PORT=<port>`` to set a default
        value.
    cleanup_on_exit : bool, optional
        Whether to shut down the connected Fluent session when PyFluent is
        exited, or the ``exit()`` method is called on the session instance,
        or if the session instance becomes unreferenced. The default is ``True``.
    start_transcript : bool, optional
        Whether to start streaming the Fluent transcript in the client. The
        default is ``True``. You can stop and start the streaming of the
        Fluent transcript subsequently via the method calls, ``start_transcript()``
        and ``stop_transcript()`` on the session object.
    show_gui : bool, optional
        Whether to display the Fluent GUI, only when ``start_instance``
        is set to ``True``. The default is ``None``, which does not
        cause the GUI to be shown. If a value of ``False`` is
        not explicitly provided, the GUI will also be shown if
        the environment variable ``PYFLUENT_SHOW_SERVER_GUI`` is set to 1.
    case_filepath : str, optional
        If provided, the case file at ``case_filepath`` is read into the Fluent session.
    case_data_filepath : str, optional
        If provided, the case and data files at ``case_data_filepath`` are read into the Fluent session.
    lightweight_mode: bool, optional
        Whether to run in lightweight mode. In lightweight mode, the lightweight settings are read into the
        current Fluent solver session. The mesh is read into a background Fluent solver session which will
        replace the current Fluent solver session once the mesh read is complete and the lightweight settings
        made by the user in the current Fluent solver session have been applied in the background Fluent
        solver session. This is all orchestrated by PyFluent and requires no special usage.
        This parameter is used only when ``case_filepath`` is provided. The default is ``False``.
    mode : str, optional
        Launch mode of Fluent to point to a specific session type.
        The default value is ``None``. Options are ``"meshing"``,
        ``"pure-meshing"`` and ``"solver"``.
    server_info_filepath: str
        Path to server-info file written out by Fluent server. The default is
        ``None``. ``server_info_filepath`` can be specified if ``start_instance``
        is ``False``, where PyFluent will use the connection information in the file to
        connect to a running Fluent session.
    password : str, optional
        Password to connect to existing Fluent instance.
    py : bool, optional
        If True, Fluent will run in Python mode. Default is None.
    gpu : bool, optional
        If True, Fluent will start with GPU Solver.
    cwd: str, Optional
        Path to specify current working directory to launch fluent from the defined directory as
        current working directory.
    topy: str or list, optional
        The string path to a Fluent journal file, or a list of such paths. Fluent will execute the
        journal(s) and write the equivalent Python journal(s).

    Returns
    -------
    Union[Meshing, PureMeshing, Solver, SolverIcing]
        Session object.

    Notes
    -----
    In job scheduler environments such as SLURM, LSF, PBS, etc... the allocated
    machines and core counts are queried from the scheduler environment and
    passed to Fluent.
    """
    if kwargs:
        if "meshing_mode" in kwargs:
            raise RuntimeError(
                "'meshing_mode' argument is no longer used."
                " Please use launch_fluent(mode='meshing') to launch in meshing mode."
            )
        else:
            raise TypeError(
                f"launch_fluent() got an unexpected keyword argument {next(iter(kwargs))}"
            )

    del kwargs
    argvals = locals()

    new_session, meshing_mode, argvals, mode = _get_session_info(argvals, mode)
    _raise_exception_g_gu_in_windows_os(additional_arguments)
    if _start_instance(start_instance):
        server_info_filepath = _get_server_info_filepath()
        launch_string = _generate_launch_string(
            argvals, meshing_mode, show_gui, additional_arguments, server_info_filepath
        )

        try:
            logger.info("Launching Fluent with cmd: %s", launch_string)
            sifile_last_mtime = Path(server_info_filepath).stat().st_mtime
            if env is None:
                env = {}
            if mode != LaunchMode.SOLVER_ICING:
                env["APP_LAUNCHED_FROM_CLIENT"] = "1"  # disables flserver datamodel
            kwargs = _get_subprocess_kwargs_for_fluent(env)
            if cwd:
                kwargs.update(cwd=cwd)
            if topy:
                launch_string += scm_to_py(topy)

            subprocess.Popen(launch_string, **kwargs)

            _await_fluent_launch(server_info_filepath, start_timeout, sifile_last_mtime)

            session = new_session.create_from_server_info_file(
                server_info_filepath=server_info_filepath,
                cleanup_on_exit=cleanup_on_exit,
                start_transcript=start_transcript,
                launcher_args=argvals,
            )
            if case_filepath:
                if meshing_mode:
                    session.tui.file.read_case(case_filepath)
                else:
                    session.read_case(case_filepath, lightweight_mode)
            if journal_filepath:
                if meshing_mode:
                    session.tui.file.read_journal(journal_filepath)
                else:
                    session.file.read_journal(journal_filepath)
            if case_data_filepath:
                if not meshing_mode:
                    session.file.read(
                        file_type="case-data", file_name=case_data_filepath
                    )
                else:
                    raise RuntimeError(
                        "Case and data file cannot be read in meshing mode."
                    )

            return session
        except Exception as ex:
            raise LaunchFluentError(launch_string) from ex
        finally:
            server_info_file = Path(server_info_filepath)
            if server_info_file.exists():
                server_info_file.unlink()
    else:
        if pypim.is_configured():
            logger.info(
                "Starting Fluent remotely. The startup configuration will be ignored."
            )

            if product_version:
                fluent_product_version = "".join(product_version.split("."))[:-1]
            else:
                fluent_product_version = "latest"

            return launch_remote_fluent(
                session_cls=new_session,
                start_timeout=start_timeout,
                start_transcript=start_transcript,
                product_version=fluent_product_version,
                cleanup_on_exit=cleanup_on_exit,
                meshing_mode=meshing_mode,
                dimensionality=version,
                launcher_args=argvals,
            )
        import ansys.fluent.core as pyfluent

        if pyfluent.BUILDING_GALLERY or os.getenv("PYFLUENT_LAUNCH_CONTAINER") == "1":
            args = _build_fluent_launch_args_string(**argvals).split()
            if meshing_mode:
                args.append(" -meshing")

            save_path = pyfluent.EXAMPLES_PATH
            # Check if save path exists
            if not os.path.exists(save_path):
                os.makedirs(save_path)

            # Assumes the container OS will be able to create the
            # EXAMPLES_PATH of host OS. With the Fluent docker
            # container, the following currently works only in linux.
            port, password = start_fluent_container(save_path, save_path, args)
            return new_session(
                fluent_connection=FluentConnection(
                    start_timeout=start_timeout,
                    port=port,
                    password=password,
                    cleanup_on_exit=cleanup_on_exit,
                    start_transcript=start_transcript,
                    launcher_args=argvals,
                )
            )
        else:
            ip, port, password = _connect_to_running_server(
                argvals, server_info_filepath
            )
            fluent_connection = FluentConnection(
                ip=ip,
                port=port,
                password=password,
                cleanup_on_exit=cleanup_on_exit,
                start_transcript=start_transcript,
                launcher_args=argvals,
            )
            new_session = _get_running_session_mode(fluent_connection, mode)
            return new_session(fluent_connection=fluent_connection)



import pytest
from util.meshing_workflow import new_mesh_session  # noqa: F401

import ansys.fluent.core as pyfluent


def test_base_session_upload(new_mesh_session):
    base_session = new_mesh_session
    with pytest.raises(AttributeError) as e_info:
        base_session._upload(
            pyfluent.EXAMPLES_PATH + "/mixing_elbow.py", "test_upload_download.py"
        )
    base_session.exit()


def test_base_session_download(new_mesh_session):
    base_session = new_mesh_session
    with pytest.raises(AttributeError) as e_info:
        base_session._download("test_upload_download.py", pyfluent.EXAMPLES_PATH)
    base_session.exit()


def test_session_upload(new_mesh_session):
    session = new_mesh_session
    with pytest.raises(AttributeError) as e_info:
        session._upload(
            pyfluent.EXAMPLES_PATH + "/mixing_elbow.py", "test_upload_download.py"
        )
    session.exit()


def test_session_download(new_mesh_session):
    session = new_mesh_session
    with pytest.raises(AttributeError) as e_info:
        session._download("test_upload_download.py", pyfluent.EXAMPLES_PATH)
    session.exit()



import difflib
from functools import partial
from typing import List


def closest_allowed_names(trial_name: str, allowed_names: str) -> List[str]:
    """Checks if the 'trail_name' is closely matching the 'allowed_names'."""
    f = partial(difflib.get_close_matches, trial_name, allowed_names)
    return f(cutoff=0.6, n=5) or f(cutoff=0.3, n=1)


def allowed_name_error_message(
    context: str, trial_name: str, allowed_values: List[str]
) -> str:
    """Provide the closest names matching the 'trial_name' from the
    'allowed_values' list."""
    message = f"{trial_name} is not an allowed {context} name.\n"
    matches = closest_allowed_names(trial_name, allowed_values)
    if matches:
        message += f"The most similar names are: {', '.join(matches)}."
    return message


def allowed_values_error(
    context: str, trial_name: str, allowed_values: List[str]
) -> str:
    return ValueError(allowed_name_error_message(context, trial_name, allowed_values))



import sys

import pytest
from util.solver_workflow import new_solver_session  # noqa: F401

import ansys.fluent.core as pyfluent
from ansys.fluent.core import examples


@pytest.mark.dev
@pytest.mark.fluent_232
@pytest.mark.skipif(
    sys.platform.startswith("linux"), reason="Linux specific issue in server"
)
def test_batch_ops_create_mesh(new_solver_session):
    solver = new_solver_session
    case_filename = examples.download_file(
        "mixing_elbow.cas.h5", "pyfluent/mixing_elbow"
    )
    with pyfluent.BatchOps(solver):
        solver.file.read_case(file_name=case_filename)
        solver.results.graphics.mesh["mesh-1"] = {}
        assert not solver.scheme_eval.scheme_eval("(case-valid?)")
        assert "mesh-1" not in solver.results.graphics.mesh.get_object_names()
    assert solver.scheme_eval.scheme_eval("(case-valid?)")
    assert "mesh-1" in solver.results.graphics.mesh.get_object_names()


@pytest.mark.dev
@pytest.mark.fluent_232
@pytest.mark.skipif(
    sys.platform.startswith("linux"), reason="Linux specific issue in server"
)
def test_batch_ops_create_mesh_and_access_fails(new_solver_session):
    solver = new_solver_session
    case_filename = examples.download_file(
        "mixing_elbow.cas.h5", "pyfluent/mixing_elbow"
    )
    with pytest.raises(KeyError):
        with pyfluent.BatchOps(solver):
            solver.file.read_case(file_name=case_filename)
            solver.results.graphics.mesh["mesh-1"] = {}
            solver.results.graphics.mesh["mesh-1"].surfaces_list = ["wall-elbow"]
    assert not solver.scheme_eval.scheme_eval("(case-valid?)")



#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import _ChildNamedObjectAccessorMixin

from ansys.fluent.core.solver.flobject import _CreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _NonCreatableNamedObjectMixin

class p_backflow_spec_gen(String):
    """
    'p_backflow_spec_gen' child.
    """

    fluent_name = "p-backflow-spec-gen"




#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import _ChildNamedObjectAccessorMixin

from ansys.fluent.core.solver.flobject import _CreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _NonCreatableNamedObjectMixin

from .turb_visc_func_mf_child import turb_visc_func_mf_child

class pb_smm_bc(NamedObject[turb_visc_func_mf_child], _NonCreatableNamedObjectMixin[turb_visc_func_mf_child]):
    """
    'pb_smm_bc' child.
    """

    fluent_name = "pb-smm-bc"

    child_object_type: turb_visc_func_mf_child = turb_visc_func_mf_child
    """
    child_object_type of pb_smm_bc.
    """



import os
from pathlib import Path
import time

import pytest

import ansys.fluent.core as pyfluent


@pytest.mark.skip(reason="Unable to read generated python journal from fluent.")
def test_single_jou(with_launching_container):
    file_path = os.path.join(pyfluent.EXAMPLES_PATH, "jou1.jou")

    with open(file_path, "w") as journal:
        journal.write('(display "from jou1.jou")')

    solver = pyfluent.launch_fluent(mode="solver", topy=file_path)
    solver.exit()

    with open(file_path) as file:
        returned = file.readlines()

    if os.path.exists(file_path):
        os.remove(file_path)

    assert returned

    gen_file_path = os.path.join(os.getcwd(), f"{Path(file_path).stem}.py")

    with open(gen_file_path) as file:
        gen_returned = file.readlines()

    assert gen_returned

    while os.path.exists(gen_file_path):
        try:
            os.remove(gen_file_path)
        except PermissionError:
            time.sleep(1)
        if not os.path.exists(gen_file_path):
            break


@pytest.mark.skip(reason="Unable to read generated python journal from fluent.")
def test_single_scm(with_launching_container):
    file_path = os.path.join(pyfluent.EXAMPLES_PATH, "jou1.scm")

    with open(file_path, "w") as journal:
        journal.write('(display "from jou1.scm")')

    solver = pyfluent.launch_fluent(mode="solver", topy=file_path)
    solver.exit()

    with open(file_path) as file:
        returned = file.readlines()

    if os.path.exists(file_path):
        os.remove(file_path)

    assert returned

    gen_file_path = os.path.join(os.getcwd(), f"{Path(file_path).stem}.py")

    with open(gen_file_path) as file:
        gen_returned = file.readlines()

    assert gen_returned

    while os.path.exists(gen_file_path):
        try:
            os.remove(gen_file_path)
        except PermissionError:
            time.sleep(1)
        if not os.path.exists(gen_file_path):
            break


@pytest.mark.skip(reason="Unable to read generated python journal from fluent.")
def test_2_jou(with_launching_container):
    file_path_1 = os.path.join(pyfluent.EXAMPLES_PATH, "jou1.jou")
    file_path_2 = os.path.join(pyfluent.EXAMPLES_PATH, "jou2.jou")

    with open(file_path_1, "w") as journal:
        journal.write('(display "from jou1.jou")')

    with open(file_path_2, "w") as journal:
        journal.write('(display "from jou2.jou")')

    solver = pyfluent.launch_fluent(mode="solver", topy=[file_path_1, file_path_2])
    solver.exit()

    with open(file_path_1) as file:
        returned1 = file.readlines()

    if os.path.exists(file_path_1):
        os.remove(file_path_1)

    assert returned1

    with open(file_path_2) as file:
        returned2 = file.readlines()

    if os.path.exists(file_path_2):
        os.remove(file_path_2)

    assert returned2

    gen_file_name = (
        Path(file_path_1).stem.split(".")[0]
        + "_"
        + Path(file_path_2).stem.split(".")[0]
        + ".py"
    )  # noqa: E501
    gen_file_path = os.path.join(os.getcwd(), gen_file_name)

    with open(gen_file_path) as file:
        gen_returned = file.readlines()

    assert gen_returned

    while os.path.exists(gen_file_path):
        try:
            os.remove(gen_file_path)
        except PermissionError:
            time.sleep(1)
        if not os.path.exists(gen_file_path):
            break


@pytest.mark.skip(reason="Unable to read generated python journal from fluent.")
def test_2_scm(with_launching_container):
    file_path_1 = os.path.join(pyfluent.EXAMPLES_PATH, "jou1.scm")
    file_path_2 = os.path.join(pyfluent.EXAMPLES_PATH, "jou2.scm")

    with open(file_path_1, "w") as journal:
        journal.write('(display "from jou1.scm")')

    with open(file_path_2, "w") as journal:
        journal.write('(display "from jou2.scm")')

    solver = pyfluent.launch_fluent(mode="solver", topy=[file_path_1, file_path_2])
    solver.exit()

    with open(file_path_1) as file:
        returned1 = file.readlines()

    if os.path.exists(file_path_1):
        os.remove(file_path_1)

    assert returned1

    with open(file_path_2) as file:
        returned2 = file.readlines()

    if os.path.exists(file_path_2):
        os.remove(file_path_2)

    assert returned2

    gen_file_name = (
        Path(file_path_1).stem.split(".")[0]
        + "_"
        + Path(file_path_2).stem.split(".")[0]
        + ".py"
    )  # noqa: E501
    gen_file_path = os.path.join(os.getcwd(), gen_file_name)

    with open(gen_file_path) as file:
        gen_returned = file.readlines()

    assert gen_returned

    while os.path.exists(gen_file_path):
        try:
            os.remove(gen_file_path)
        except PermissionError:
            time.sleep(1)
        if not os.path.exists(gen_file_path):
            break



#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import _ChildNamedObjectAccessorMixin

from ansys.fluent.core.solver.flobject import _CreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _NonCreatableNamedObjectMixin

class kw_low_re_correction(Boolean):
    """
    Enable/disable the k-omega low Re option.
    """

    fluent_name = "kw-low-re-correction"

